/*
 * ClauseDatabase.h
 *
 *  Created on: Nov 12, 2020
 *      Author: kokoro
 */

#ifndef CLEANSOLVER_CORESOLVER_CLAUSEDATABASE_H_
#define CLEANSOLVER_CORESOLVER_CLAUSEDATABASE_H_

#include "../../Utilities/PrintUtilities/DratPrinter/DratPrinter.h"
#include "../../Utilities/BooleanTypes/SolverTypes.h"
#include "../../Utilities/Containers/Vec.h"
#include "ImplicationGraph.h"
#include "VariableDatabase.h"

namespace CleanMaple {

/*
 * TODO: Get rid of the following Macros.
 */
#ifndef LOCAL
#define LOCAL 0
#define TIER2 2
#define CORE  3
#endif

class ClauseDatabase {
public:
	// TODO: 	All of the following need to be private. The only purpose of
	//			their publicity is to have compiling versions while being in the
	// 			middle of the process of factoring all that belongs to this
	//			class out.
	Minisat::ClauseAllocator ca;
	Minisat::vec<Minisat::CRef> clauses;          // List of problem clauses.
	Minisat::vec<Minisat::CRef> learnts_core;     // List of learnt clauses.
	Minisat::vec<Minisat::CRef> learnts_tier2;    // List of learnt clauses.
	Minisat::vec<Minisat::CRef> learnts_local;    // List of learnt clauses.
	double garbage_frac; // The fraction of wasted memory allowed before a garbage collection is triggered.

	/*
	 * TODO: All of the following is supposed to be in a seperate class. however
	 * this is currently not possible, because the WatcherDeleted need the
	 * ClauseAllocator, while the ClauseDatabase needs the WatcherScheme as a
	 * member.
	 */
	struct Watcher {
		Minisat::CRef cref;
		Minisat::Lit blocker;
		Watcher(Minisat::CRef cr, Minisat::Lit p) :
				cref(cr), blocker(p) {
		}
		bool operator==(const Watcher& w) const {
			return cref == w.cref;
		}
		bool operator!=(const Watcher& w) const {
			return cref != w.cref;
		}
	};

	struct WatcherDeleted {
		const Minisat::ClauseAllocator &ca;
		WatcherDeleted(const Minisat::ClauseAllocator& _ca) :
				ca(_ca) {
		}
		bool operator()(const Watcher& w) const {
			return ca[w.cref].mark() == 1;
		}
	};


	/*
	 *  Watches for binary clauses only.
	 */
	Minisat::OccLists<Minisat::Lit, Minisat::vec<Watcher>, WatcherDeleted> watches_bin;

	/*
	 * 'watches[lit]' is a list of constraints watching 'lit' (will go there if
	 * literal becomes true).
	 */
	Minisat::OccLists<Minisat::Lit, Minisat::vec<Watcher>, WatcherDeleted> watches;

	Minisat::OccLists<Minisat::Lit, Minisat::vec<Watcher>, WatcherDeleted> watches_local;

	Minisat::OccLists<Minisat::Lit, Minisat::vec<Watcher>, WatcherDeleted> watches_high_ranked;


	bool remove_satisfied; // Indicates whether possibly inefficient linear scan for satisfied clauses should be performed in 'simplify'.
	int core_lbd_cut;

private:

	const VariableDatabase &variable_database;

	/*
	 * TODO: Make the following implication_graph const, which is currently not
	 * possible due to needed access to the set_reason() function within the
	 * removeClause() member function of the clause_database.
	 */
	/* TODO: const */ ImplicationGraph &implication_graph;
	/* TODO: const */ DratPrinter &drat_printer;
	uint64_t clauses_literals, learnts_literals;
	double cla_inc;          // Amount to bump next clause with.
	double clause_decay;

public:
	unsigned max_rank;

public:
	/*
	 * TODO: The Implication_graph is supposed to be const, which currently is
	 * not possible. Change this.
	 */
	ClauseDatabase(const VariableDatabase& a_variable_database, /* const */ ImplicationGraph& a_implication_graph, /* const */ DratPrinter& a_drat_printer ) :
					ca(),
					clauses(),
					learnts_core(),
					learnts_tier2(),
					learnts_local(),
					watches_bin(WatcherDeleted(ca)),
					watches(WatcherDeleted(ca)),
					watches_local(WatcherDeleted(ca)),
					watches_high_ranked(WatcherDeleted(ca)),
					remove_satisfied(true),
					core_lbd_cut(3),
					variable_database(a_variable_database),
					implication_graph(a_implication_graph),
					drat_printer(a_drat_printer),
					garbage_frac(0.20),
					clauses_literals(0),
					learnts_literals(0),
					cla_inc(1),
					clause_decay(0.999),
					max_rank(0),
					age_to_reduce_local_watchers(500){
		// Do nothing.
	}
	;
	virtual ~ClauseDatabase();
	ClauseDatabase(const ClauseDatabase& other) = delete;
	ClauseDatabase(ClauseDatabase&& other) = delete;
	ClauseDatabase& operator=(const ClauseDatabase& other) = delete;
	ClauseDatabase& operator=(ClauseDatabase&& other) = delete;

	// TODO: Get rid of this parameter new_ca. See implementation.
	void reallocate(Minisat::ClauseAllocator& new_ca);

	void removeSatisfied();

	/*
	 * TODO: Check whether the following is true:
	 * 		"Removes half of the learnt clauses, minus the clauses locked by the
	 * 		current assignment. Locked clauses are clauses that are reason to
	 * 		some assignment. Binary clauses are never removed."
	 */
	void reduceDB();
	void reduceDB_Tier2(uint64_t& current_number_of_conflicts);

	int age_to_reduce_local_watchers;
	void update_watchers_of_learnts_core(uint64_t current_n_of_conflicts);
	void update_watchers_of_original_clauses(uint64_t current_n_of_conflicts);
	void update_watchers_of_learnts_local();
	void update_watchers_of_learnts_tier2(uint64_t current_n_of_conflicts);


	inline void claBumpActivity (Minisat::Clause& c);
	inline void claDecayActivity();

	/*
	 * TODO: Make this private. The only reason this is public are the
	 * simplify_learnt-methods from the Solver and strengthenClause from the
	 * SimpSolver.
	 */
	inline void detach_clause_from_watcher_scheme(Minisat::CRef cr, bool strict = false);

	/*
	 * TODO: Make this function private. This currently can not be achieved,
	 * because strengthen clause, and the simplifyLearnts-methods need access to
	 * this function. Also the Solver::addClause_-method calls this function,
	 * which is something that needs to be changed as well.
	 */
	inline void attach_clause_to_watcher_scheme(Minisat::CRef cr);

	/*
	 * TODO: Make this function private. This currently can not be achieved,
	 * because strengthen clause, and the simplifyLearnts-methods need access to
	 * this function.
	 * TODO: Implement hint Clause if calculated earlier as a performance gain.
	 */
	inline void remove_clause_from_database(const Minisat::CRef& cr);

	/*
	 * TODO: Implement learn unary clause, even though it will be difficult,
	 * because learning a unary clause is modeled through the implication graph.
	 */
	inline Minisat::CRef learn_non_unary_clause(Minisat::vec<Minisat::Lit> &c,
			int &lbd_from_search_method,
			int &rank_estimate_from_search_method,
			const uint64_t &current_number_of_conflicts);

	/*
	 * TODO: What exactly is the following, and why do we need it? See comment
	 * where it is called. Should be removed as soon as possible.
	 * TODO: This method should not even exist.
	 *
	 */
	uint64_t calculate_simpDB_props() const;

	/*
	 * TODO: Make this function private. This currently can not be achieved,
	 * because strengthen clause, and the simplifyLearnts-methods need access to
	 * this function. Also the Solver::addClause_-method calls this function,
	 * which is something that needs to be changed as well.
	 * TODO: This method should not even exist.
	 */
	inline void update_literal_counters_add(const Minisat::Clause& c);

	/*
	 * TODO: Make this function private. This currently can not be achieved,
	 * because strengthen clause, and the simplifyLearnts-methods need acces to
	 * this function.
	 * TODO: This method should not even exist.
	 */
	inline void update_literal_counters_substract(const Minisat::Clause &c);

	/*
	 * TODO: This function should not even exist. Take a look at the
	 * implementation of removeClause.
	 *  	"Returns TRUE if the clause is the reason for some implication in the
	 *  	current state."
	 * TODO: Make this function private.
	 */
	inline bool locked(const Minisat::Clause& c) const;

private:

	inline void put_clause_into_corresponding_tier(const Minisat::CRef &cr,
			const uint64_t &current_number_of_conflicts);

	/*
	 * TODO: This function should not even exist. Take a look at the
	 * implementation of removeClause.
	 */
	inline void unlock(const Minisat::Clause& c);


	void reallocate_all_watches(Minisat::ClauseAllocator& new_ca);
	void reallocate_all_reason_clauses(Minisat::ClauseAllocator& new_ca);

	void reallocate_all_clauses(Minisat::ClauseAllocator& new_ca);
	void design_fixing_unlock_if_locked(Minisat::Clause& c);
	void removeSatisfied(Minisat::vec<Minisat::CRef>& cs);
	void safeRemoveSatisfied(Minisat::vec<Minisat::CRef>& cs, unsigned valid_mark);
	inline void rescale_clause_activities();
};

/*
 * Implementation of inline functions:
 */



/*
 * TODO: This function should not even exist. Take a look at the
 * implementation of removeClause.
 * TODO: Make this function private.
 * TODO: This implementation relies on the fact, that the implied literal is in
 * position c[0] for non-binary clauses.
 */
inline bool ClauseDatabase::locked(const Minisat::Clause& c) const {
	// TODO: Using directive for the macros
	using namespace Minisat;
	int i = c.size() != 2 ?
			0 : (variable_database.value(c[0]) == l_True ? 0 : 1);
	return variable_database.value(c[i]) == l_True
			&& implication_graph.reason(var(c[i])) != CRef_Undef
			&& ca.lea(implication_graph.reason(var(c[i]))) == &c;
}


inline void ClauseDatabase::unlock(const Minisat::Clause& c) {
	// TODO: Move all of the following to the implication_graph,
	// possibly including the above if(locked(c) { ... }
	using namespace Minisat;
	Minisat::Lit implied =
			c.size() != 2 ?
					c[0] :
					(variable_database.value(c[0]) == l_True ? c[0] : c[1]);
	implication_graph.set_reason(var(implied), Minisat::CRef_Undef);

}

/*
 * TODO: The following two functions need to be called from a not yet existing
 * addClause() member-function of the ClauseDatabase. See implementation of
 * the functions, which call the following function.
 */
inline void ClauseDatabase::update_literal_counters_add(const Minisat::Clause& c) {
	if (c.learnt())
		learnts_literals += c.size();
	else
		clauses_literals += c.size();
}

inline void ClauseDatabase::update_literal_counters_substract(const Minisat::Clause &c) {
	if (c.learnt())
		learnts_literals -= c.size();
	else
		clauses_literals -= c.size();
}


/*
 * TODO: What exactly is the following, and why do we need it? See comment
 * where it is called. Should be removed as soon as possible.
 */
inline uint64_t ClauseDatabase::calculate_simpDB_props() const {
	return clauses_literals + learnts_literals;
}


/*
 * The following function can not be put here, because the removeClause is not
 * yet properly implemented as a member of this class. In it's current form, the
 * removeClause misses some stuff which can not yet be moved inside from the
 * corresponding function from the solver.
 */
//void ClauseDatabase::reduceDB() {
// TODO: Implement this function.
//}
inline void ClauseDatabase::remove_clause_from_database(const Minisat::CRef& cr) {
	Minisat::Clause &c = ca[cr];
	drat_printer.removeClause(c);
	detach_clause_from_watcher_scheme(cr);
	update_literal_counters_substract(c);
	design_fixing_unlock_if_locked(c);
	/*
	 * FIXME:
	 * Can we get more clarity, what is happening in the following two
	 * lines? I' mean the ClauseAllocator free's the space, however,
	 * existing CRef's pointing to the clause are still able to see, that
	 * the clause has been deleted. What exactly is happening here? Also
	 * I feel like the marking of the clause then needs to be inside of
	 * the free function. And it must somehow be entirely cleanly made clear
	 * that CRef's can point to invalid, because deleted clauses. I.e. each
	 * function using CRef's would need to forced to take care of this by
	 * design.
	 */
	c.mark(1);
	ca.free(cr);
}

inline void ClauseDatabase::attach_clause_to_watcher_scheme(Minisat::CRef cr) {
	const Minisat::Clause &c = ca[cr];
	assert(c.size() > 1);
	if (c.size() == 2) {
		assert(c.watcher_mark() == 2);
	}
	Minisat::OccLists<Minisat::Lit,
			Minisat::vec<CleanMaple::ClauseDatabase::Watcher>,
			CleanMaple::ClauseDatabase::WatcherDeleted> &ws =
			c.size() == 2 ? watches_bin :
			c.watcher_mark() == 0 ?
					watches :
					c.watcher_mark() == 1 ? watches_high_ranked : watches_local;
	ws[~c[0]].push(CleanMaple::ClauseDatabase::Watcher(cr, c[1]));
	ws[~c[1]].push(CleanMaple::ClauseDatabase::Watcher(cr, c[0]));
}


/*
 * TODO: Have a second method, where the clause c is passed as reference as well
 * in order to speed things up. The reason is, that the clause c is at least in
 * remove_clause_from_database() which calls this function, already calculated.
 */
inline void ClauseDatabase::detach_clause_from_watcher_scheme(Minisat::CRef cr, bool strict) {
	const Minisat::Clause &c = ca[cr];
	assert(c.size() > 1);
	assert(c.size() != 2 || c.watcher_mark() == 2);
	Minisat::OccLists<Minisat::Lit,
			Minisat::vec<CleanMaple::ClauseDatabase::Watcher>,
			CleanMaple::ClauseDatabase::WatcherDeleted> &ws =
			c.size() == 2 ? watches_bin :
			c.watcher_mark() == 0 ?
					watches :
					c.watcher_mark() == 1 ? watches_high_ranked : watches_local;

	if (strict) {
		remove(ws[~c[0]], CleanMaple::ClauseDatabase::Watcher(cr, c[1]));
		remove(ws[~c[1]], CleanMaple::ClauseDatabase::Watcher(cr, c[0]));
	} else {
		// Lazy detaching: (NOTE! Must clean all watcher lists before garbage collecting this clause)
		ws.smudge(~c[0]);
		ws.smudge(~c[1]);
	}
}

inline void ClauseDatabase::claDecayActivity() {
	 cla_inc *= (1 / clause_decay);
}

inline Minisat::CRef ClauseDatabase::learn_non_unary_clause(
		Minisat::vec<Minisat::Lit> &learnt_clause_from_search_method,
		int &lbd_from_search_method,
		int &rank_estimate_from_search_method,
		const uint64_t &current_number_of_conflicts) {
	Minisat::CRef cr = ca.alloc(learnt_clause_from_search_method, true);
	ca[cr].set_lbd(lbd_from_search_method);
	ca[cr].set_rank_estimate(rank_estimate_from_search_method);
	if (rank_estimate_from_search_method > max_rank)
		max_rank = rank_estimate_from_search_method;
	put_clause_into_corresponding_tier(cr, current_number_of_conflicts);
	if (ca[cr].size() != 2) {
		if (false && rank_estimate_from_search_method > (0.95 * max_rank)) {
			// disabled
			ca[cr].watcher_mark(1); // 1 = high ranked
		} else {
			ca[cr].watcher_mark(3); // 3 = local
		}
	} else
	{
		ca[cr].watcher_mark(2); // 2 = binary
	}
	attach_clause_to_watcher_scheme(cr);
    update_literal_counters_add(ca[cr]);
	return cr;
}

inline void ClauseDatabase::put_clause_into_corresponding_tier(
		const Minisat::CRef &cr, const uint64_t &current_number_of_conflicts) {
	if (ca[cr].lbd() <= core_lbd_cut) {
		learnts_core.push(cr);
		ca[cr].mark(CORE);
		ca[cr].touched() = current_number_of_conflicts;
	} else if (ca[cr].lbd() <= 6) {
		learnts_tier2.push(cr);
		ca[cr].mark(TIER2);
		ca[cr].touched() = current_number_of_conflicts;
	} else {
		learnts_local.push(cr);
		claBumpActivity(ca[cr]);
	}
}

inline void ClauseDatabase::rescale_clause_activities() {
	for (int i = 0; i < learnts_local.size(); i++)
		ca[learnts_local[i]].activity() *= 1e-20;
	cla_inc *= 1e-20;
}

/*
 * Originally had the following comment:
 * 		"Increase a clause with the current 'bump' value."
 */
inline void ClauseDatabase::claBumpActivity(Minisat::Clause &c) {
	if ((c.activity() += cla_inc) > 1e20) {
		rescale_clause_activities();
	}
}

} /* namespace CleanMaple */

#endif /* CLEANSOLVER_CORESOLVER_CLAUSEDATABASE_H_ */
