/*
 * Propagator.h
 *
 *  Created on: Mar 7, 2021
 *      Author: kokoro
 */

#ifndef CLEANSOLVER_CORESOLVER_PROPAGATOR_H_
#define CLEANSOLVER_CORESOLVER_PROPAGATOR_H_

#include "../../Utilities/BooleanTypes/SolverTypes.h"

#include "ClauseDatabase.h"
#include "Brancher.h"

namespace CleanMaple {
using namespace Minisat;
class Propagator {

public:
	Propagator(ClauseDatabase& a_clause_database,
			VariableDatabase& a_variable_database,
			ImplicationGraph& a_implication_graph, Brancher& a_brancher) :
					clause_database(a_clause_database),
					variable_database(a_variable_database),
					implication_graph(a_implication_graph),
					brancher(a_brancher),
					propagations(0),
					simpDB_props(0) {
	}
	;
	virtual ~Propagator() {};
	Propagator(const Propagator& other) = delete;
	Propagator(Propagator&& other) = delete;
	Propagator& operator=(const Propagator& other) = delete;
	Propagator& operator=(Propagator&& other) = delete;

	Minisat::CRef propagate();
	inline void  uncheckedEnqueue(Minisat::Lit p, int level = 0, Minisat::CRef from = Minisat::CRef_Undef);

	uint64_t propagations;
    int64_t  simpDB_props;     // Remaining number of propagations that must be made before next execution of 'simplify()'.


private:
	ClauseDatabase &clause_database;
	VariableDatabase &variable_database;
	ImplicationGraph &implication_graph;
	Brancher &brancher;

	inline void clean_up_before_propagate();
	inline void find_MaxLevel_at_Unit_Propagation(Minisat::Clause& c, int& nMaxLevel,
			int& nMaxInd);
	inline void  make_second_lit_have_highest_level(Minisat::Clause& c,
			const int& nMaxInd, CleanMaple::ClauseDatabase::Watcher*& j,
			CleanMaple::ClauseDatabase::Watcher& w);
	inline void  handle_unit_assignment_during_propagation(int& currLevel,
			Minisat::Lit& first, Minisat::CRef& cr, Minisat::Clause& c,
			CleanMaple::ClauseDatabase::Watcher*& j,
			CleanMaple::ClauseDatabase::Watcher& w);
	inline void  copy_remaining_watchers(CleanMaple::ClauseDatabase::Watcher*& i,
			CleanMaple::ClauseDatabase::Watcher*& j,
			CleanMaple::ClauseDatabase::Watcher*& end);
	inline Minisat::Lit& move_false_lit_to_second_position(Minisat::Clause& cr,
			Minisat::Lit& p);
	inline void  set_new_watcher(Minisat::Clause& c, int& k, Minisat::Lit& false_lit,
			CleanMaple::ClauseDatabase::Watcher& w);
	inline bool search_and_set_new_watcher(Minisat::Clause& c, Minisat::Lit& false_lit,
			CleanMaple::ClauseDatabase::Watcher& w);
	inline Minisat::CRef& propagate_binary_clauses(Minisat::Lit& p, int& currLevel,
			Minisat::CRef& confl);
	inline Minisat::CRef& propagate_remaining_clauses(Minisat::Lit& p, int& currLevel,
			Minisat::CRef& confl);
	inline void  propagate_lit_on_clause(Minisat::Lit& p, int& currLevel,
			Minisat::CRef& confl, CleanMaple::ClauseDatabase::Watcher*& i,
			CleanMaple::ClauseDatabase::Watcher*& j,
			CleanMaple::ClauseDatabase::Watcher*& end);
};




inline void  Propagator::clean_up_before_propagate() {
    clause_database.watches.cleanAll();
    clause_database.watches_bin.cleanAll();
}



inline void  Propagator::uncheckedEnqueue(Minisat::Lit p, int level, Minisat::CRef from)
{
    assert(variable_database.value(p) == l_Undef);
    Var x = var(p);
    brancher.uncheckedEnque_adjustment(x);
    variable_database.set_value(x, lbool(!sign(p)));
    implication_graph.set_reason(x, from, level);
    implication_graph.trail.push_(p);
}

inline void  Propagator::find_MaxLevel_at_Unit_Propagation(Minisat::Clause& c,
		int& nMaxLevel, int& nMaxInd) {
	// pass over all the literals in the clause and find the one with the biggest level
	for (int nInd = 2; nInd < c.size(); ++nInd) {
		int nLevel = implication_graph.level(var(c[nInd]));
		if (nLevel > nMaxLevel) {
			nMaxLevel = nLevel;
			nMaxInd = nInd;
		}
	}
}

inline void  Propagator::make_second_lit_have_highest_level(Clause& c,
		const int& nMaxInd, CleanMaple::ClauseDatabase::Watcher*& j,
		CleanMaple::ClauseDatabase::Watcher& w) {
	if (nMaxInd != 1) {
		std::swap(c[1], c[nMaxInd]);
		*j--; // undo last watch
		clause_database.watches[~c[1]].push(w);
	}
}

inline void  Propagator::handle_unit_assignment_during_propagation(int& currLevel,
		Lit& first, CRef& cr, Clause& c,
		CleanMaple::ClauseDatabase::Watcher*& j,
		CleanMaple::ClauseDatabase::Watcher& w) {
	if (currLevel == implication_graph.decisionLevel()) {
		uncheckedEnqueue(first, currLevel, cr);
	}
	else
	{
		// We need to find the correct level:
		int nMaxLevel = currLevel;
		int nMaxInd = 1;
		find_MaxLevel_at_Unit_Propagation(c,nMaxLevel,nMaxInd);
		make_second_lit_have_highest_level(c, nMaxInd, j, w);
		uncheckedEnqueue(first, nMaxLevel, cr);
	}
}

inline void  Propagator::copy_remaining_watchers(
		CleanMaple::ClauseDatabase::Watcher*& i,
		CleanMaple::ClauseDatabase::Watcher*& j,
		CleanMaple::ClauseDatabase::Watcher*& end) {
	// Copy the remaining watches:
	i++;
	while (i < end) {
		*j++ = *i;
		i++;
	}
}

inline Minisat::Lit& Propagator::move_false_lit_to_second_position(Minisat::Clause& c, Minisat::Lit& false_lit) {
    if (c[0] == false_lit)
        c[0] = c[1], c[1] = false_lit;
    //assert(c[1] == false_lit);
    return c[0];
}

inline void  Propagator::set_new_watcher(Minisat::Clause& c, int& k,
		Minisat::Lit& false_lit, CleanMaple::ClauseDatabase::Watcher& w) {
	c[1] = c[k];
	c[k] = false_lit;
	clause_database.watches[~c[1]].push(w);
}

inline bool Propagator::search_and_set_new_watcher(Minisat::Clause& c,
		Minisat::Lit& false_lit, CleanMaple::ClauseDatabase::Watcher& w) {
	for (int k = 2; k < c.size(); k++) {
		if (variable_database.value(c[k]) != l_False) {
			c[1] = c[k];
			c[k] = false_lit;
			clause_database.watches[~c[1]].push(w);
			return true;
		}
	}
	return false;
}

inline Minisat::CRef& Propagator::propagate_binary_clauses(Minisat::Lit& p, int& currLevel, Minisat::CRef& confl) {
	Minisat::vec<CleanMaple::ClauseDatabase::Watcher> &ws_bin =
			clause_database.watches_bin[p]; // Propagate binary clauses first.
	for (int k = 0; k < ws_bin.size(); k++) {
		Minisat::Lit the_other = ws_bin[k].blocker;
		if (variable_database.value(the_other) == l_False) {
			confl = ws_bin[k].cref;
			return confl;
		} else if (variable_database.value(the_other) == l_Undef) {
			uncheckedEnqueue(the_other, currLevel, ws_bin[k].cref);
		}
	}
	return confl;
}

inline void  Propagator::propagate_lit_on_clause(Minisat::Lit& p, int& currLevel,
		Minisat::CRef& confl, CleanMaple::ClauseDatabase::Watcher*& i,
		CleanMaple::ClauseDatabase::Watcher*& j,
		CleanMaple::ClauseDatabase::Watcher*& end) {
	Minisat::CRef cr = i->cref;
	Lit blocker = i->blocker;
	Clause &c = clause_database.ca[cr];
	Lit false_lit = ~p;
	Lit first_lit = move_false_lit_to_second_position(c, false_lit);
	CleanMaple::ClauseDatabase::Watcher w =
			CleanMaple::ClauseDatabase::Watcher(cr, first_lit);

	if (first_lit
			!= blocker && variable_database.value(first_lit) == l_True) {
		*j++ = w;
	} else if (!search_and_set_new_watcher(c, false_lit, w)) {
		*j++ = w;
		if (variable_database.value(first_lit) == l_False) {
			confl = cr;
			implication_graph.qhead = implication_graph.trail.size();
			copy_remaining_watchers(i, j, end);
		} else {
			handle_unit_assignment_during_propagation(currLevel,
					first_lit, cr, c, j, w);
		}
	}
}

inline Minisat::CRef& Propagator::propagate_remaining_clauses(Minisat::Lit& p,
		int& currLevel, Minisat::CRef& confl) {
	Minisat::vec<CleanMaple::ClauseDatabase::Watcher> &ws = clause_database
			.watches[p];
	CleanMaple::ClauseDatabase::Watcher *j, *end;
	j = (CleanMaple::ClauseDatabase::Watcher*) ws;
	end = j + ws.size();

	// If watcher at position i needs to be kept, we will copy it to position
	// j until the end.
	for (CleanMaple::ClauseDatabase::Watcher *i =
			(CleanMaple::ClauseDatabase::Watcher*) ws; i < end; ++i) {
		// Try to avoid inspecting the clause:
		Lit blocker = i->blocker;
		if (variable_database.value(blocker) == l_True) {
			*j++ = *i; // i.e. keep watcher
		} else {
			propagate_lit_on_clause(p, currLevel, confl, i, j, end);
		}
	}
	ws.shrink(end - j);
}


/*_________________________________________________________________________________________________
|
|  propagate : [void]  ->  [Clause*]
|
|  Description:
|    Propagates all enqueued facts. If a conflict arises, the conflicting clause is returned,
|    otherwise CRef_Undef.
|
|    Post-conditions:
|      * the propagation queue is empty, even if there was a conflict.
|________________________________________________________________________________________________@*/
inline Minisat::CRef Propagator::propagate()
{
    clean_up_before_propagate();

    Minisat::CRef    confl     = CRef_Undef;
    int     num_props = 0;

	while (implication_graph.qhead < implication_graph.trail.size()) {
		num_props++;
		Minisat::Lit p = implication_graph.trail[implication_graph.qhead++];
		int currLevel = implication_graph.level(var(p));


		if ( propagate_binary_clauses(p,currLevel,confl) != CRef_Undef ) {
			return confl;
		}

		propagate_remaining_clauses(p,currLevel,confl);
    }
    propagations += num_props;
    simpDB_props -= num_props;

    return confl;
}


} /* namespace CleanMaple */

#endif /* CLEANSOLVER_CORESOLVER_PROPAGATOR_H_ */
