/***************************************************************************************[Solver.cc]
MiniSat -- Copyright (c) 2003-2006, Niklas Een, Niklas Sorensson
           Copyright (c) 2007-2010, Niklas Sorensson
 
Chanseok Oh's MiniSat Patch Series -- Copyright (c) 2015, Chanseok Oh
 
Maple_LCM, Based on MapleCOMSPS_DRUP -- Copyright (c) 2017, Mao Luo, Chu-Min LI, Fan Xiao: implementing a learnt clause minimisation approach
Reference: M. Luo, C.-M. Li, F. Xiao, F. Manya, and Z. L. , “An effective learnt clause minimization approach for cdcl sat solvers,” in IJCAI-2017, 2017, pp. to–appear.

Maple_LCM_Dist, Based on Maple_LCM -- Copyright (c) 2017, Fan Xiao, Chu-Min LI, Mao Luo: using a new branching heuristic called Distance at the beginning of search
 
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/

#include "Solver.h"

#include <math.h>
#include <algorithm>
#include <signal.h>
#include <unistd.h>

#include "../../Utilities/Algorithms/Sort.h"
#include "../../Utilities/System/System.h"

using namespace Minisat;

//#define PRINT_OUT



//=================================================================================================
// Options:


static const char* _cat = "CORE";

static IntOption     opt_ccmin_mode        (_cat, "ccmin-mode",  "Controls conflict clause minimization (0=none, 1=basic, 2=deep)", 2, IntRange(0, 2));
static IntOption     opt_phase_saving      (_cat, "phase-saving", "Controls the level of phase saving (0=none, 1=limited, 2=full)", 2, IntRange(0, 2));
static IntOption     opt_restart_first     (_cat, "rfirst",      "The base restart interval", 100, IntRange(1, INT32_MAX));
static DoubleOption  opt_restart_inc       (_cat, "rinc",        "Restart interval increase factor", 2, DoubleRange(1, false, HUGE_VAL, false));
static IntOption     opt_chrono            (_cat, "chrono",  "Controls if to perform chrono backtrack", 100, IntRange(-1, INT32_MAX));
static IntOption     opt_conf_to_chrono    (_cat, "confl-to-chrono",  "Controls number of conflicts to perform chrono backtrack", 4000, IntRange(-1, INT32_MAX));


//=================================================================================================
// Constructor/Destructor:


Solver::Solver() :

    // Parameters (user settable):
    //
    drat_printer()
  , verbosity        (0)
  , ccmin_mode       (opt_ccmin_mode)
  , phase_saving     (opt_phase_saving)
  , restart_first    (opt_restart_first)
  , restart_inc      (opt_restart_inc)

  // Statistics: (formerly in 'SolverStats')
  //
  , starts(0), decisions(0), conflicts(0), conflicts_VSIDS(0)
  , max_literals(0), tot_literals(0)
  , chrono_backtrack(0), non_chrono_backtrack(0)

  , ok                 (true),

variable_database(),
brancher(variable_database,conflicts),
implication_graph(),
clause_database(variable_database, implication_graph, drat_printer),
propagator(clause_database,
		variable_database,
		implication_graph,
		brancher)

  , global_lbd_sum     (0)
  , lbd_queue          (50)
  , confl_to_chrono    (opt_conf_to_chrono)
  , chrono			   (opt_chrono)
  
  , counter            (0)

{}


Solver::~Solver()
{
}


// simplify All
//
CRef Solver::simplePropagate()
{
    CRef    confl = CRef_Undef;
    int     num_props = 0;
    clause_database.watches.cleanAll();
    clause_database.watches_bin.cleanAll();
    clause_database.watches_local.cleanAll();
	clause_database.watches_high_ranked.cleanAll();
	while (implication_graph.qhead < implication_graph.trail.size()) {
		{
			// First normal_watches
			Lit p = implication_graph.trail[implication_graph.qhead++]; // 'p' is enqueued fact to propagate.
			vec<CleanMaple::ClauseDatabase::Watcher> &ws =
					clause_database.watches[p];
			CleanMaple::ClauseDatabase::Watcher *i, *j, *end;
			num_props++;

			// First, Propagate binary clauses
			vec<CleanMaple::ClauseDatabase::Watcher> &wbin =
					clause_database.watches_bin[p];

			for (int k = 0; k < wbin.size(); k++) {

				Lit imp = wbin[k].blocker;

				if (variable_database.value(imp) == l_False) {
					return wbin[k].cref;
				}

				if (variable_database.value(imp) == l_Undef) {
					simpleUncheckEnqueue(imp, wbin[k].cref);
				}
			}

			for (i = j = (CleanMaple::ClauseDatabase::Watcher*) ws, end = i
					+ ws.size(); i != end;) {
				// Try to avoid inspecting the clause:
				Lit blocker = i->blocker;
				if (variable_database.value(blocker) == l_True) {
					*j++ = *i++;
					continue;
				}

				// Make sure the false literal is data[1]:
				CRef cr = i->cref;
				Clause &c = clause_database.ca[cr];
				Lit false_lit = ~p;
				if (c[0] == false_lit)
					c[0] = c[1], c[1] = false_lit;
				assert(c[1] == false_lit);
				//  i++;

				// If 0th watch is true, then clause is already satisfied.
				// However, 0th watch is not the blocker, make it blocker using a new watcher w
				// why not simply do i->blocker=first in this case?
				Lit first = c[0];
				//  CleanMaple::ClauseDatabase::Watcher w     = CleanMaple::ClauseDatabase::Watcher(cr, first);
				if (first != blocker && variable_database.value(first) == l_True) {
					i->blocker = first;
					*j++ = *i++;
					continue;
				}

				else {
					for (int k = 2; k < c.size(); k++) {

						if (variable_database.value(c[k]) != l_False) {
							// watcher i is abandonned using i++, because cr watches now ~c[k] instead of p
							// the blocker is first in the watcher. However,
							// the blocker in the corresponding watcher in ~first is not c[1]
							CleanMaple::ClauseDatabase::Watcher w =
									CleanMaple::ClauseDatabase::Watcher(cr,
											first);
							i++;
							c[1] = c[k];
							c[k] = false_lit;
							clause_database.watches[~c[1]].push(w);
							goto simplePropagateNextClause;
						}
					}
				}

				// Did not find watch -- clause is unit under assignment:
				i->blocker = first;
				*j++ = *i++;
				if (variable_database.value(first) == l_False) {
					confl = cr;
					implication_graph.qhead = implication_graph.trail.size();
					// Copy the remaining watches:
					while (i < end)
						*j++ = *i++;
				} else {
					simpleUncheckEnqueue(first, cr);
				}
				simplePropagateNextClause: ;
			}
			ws.shrink(i - j);
		}
		{
			// Next local and high_ranked watches... TODO!
		}
	}

	return confl;
}

void Solver::simpleUncheckEnqueue(Lit p, CRef from){
    assert(variable_database.value(p) == l_Undef);
    variable_database.set_value(var(p),lbool(!sign(p))); // this makes a lbool object whose value is sign(p)
    // FIXME: Why does the following not set the level?
    implication_graph.set_reason(var(p),from);
    implication_graph.trail.push_(p);
}

void Solver::cancelUntilTrailRecord()
{
    for (int c = implication_graph.trail.size() - 1; c >= trailRecord; c--)
    {
        Var x = var(implication_graph.trail[c]);
        variable_database.set_value(x,l_Undef);

    }
    implication_graph.qhead = trailRecord;
    implication_graph.trail.shrink(implication_graph.trail.size() - trailRecord);

}

void Solver::litsEnqueue(int cutP, Clause& c)
{
    for (int i = cutP; i < c.size(); i++)
    {
        simpleUncheckEnqueue(~c[i]);
    }
}

bool Solver::removed(CRef cr) {
    return clause_database.ca[cr].mark() == 1;
}

void Solver::simpleAnalyze(CRef confl, vec<Lit>& out_learnt, vec<CRef>& reason_clause, bool True_confl)
{
    int pathC = 0;
    Lit p = lit_Undef;
    int index = implication_graph.trail.size() - 1;

    do{
        if (confl != CRef_Undef){
            reason_clause.push(confl);
            Clause& c = clause_database.ca[confl];
            // Special case for binary clauses
            // The first one has to be SAT
            if (p != lit_Undef && c.size() == 2 && variable_database.value(c[0]) == l_False) {

                assert(variable_database.value(c[1]) == l_True);
                Lit tmp = c[0];
                c[0] = c[1], c[1] = tmp;
            }
            // if True_confl==true, then choose p begin with the 1th index of c;
            for (int j = (p == lit_Undef && True_confl == false) ? 0 : 1; j < c.size(); j++){
                Lit q = c[j];
                if (!seen[var(q)]){
                    seen[var(q)] = 1;
                    pathC++;
                }
            }
        }
        else if (confl == CRef_Undef){
            out_learnt.push(~p);
        }
        // if not break, while() will come to the index of trail blow 0, and fatal error occur;
        if (pathC == 0) break;
        // Select next clause to look at:
        while (!seen[var(implication_graph.trail[index--])]);
        // if the reason cr from the 0-level assigned var, we must break avoid move forth further;
        // but attention that maybe seen[x]=1 and never be clear. However makes no matter;
        if (trailRecord > index + 1) break;
        p = implication_graph.trail[index + 1];
        confl = implication_graph.reason(var(p));
        seen[var(p)] = 0;
        pathC--;

    } while (pathC >= 0);
}

void Solver::simplifyLearnt(Clause& c)
{
    trailRecord = implication_graph.trail.size();// record the start pointer

    vec<Lit> falseLit;
    falseLit.clear();

    //sort(&c[0], c.size(), VarOrderLevelLt(vardata));

    bool True_confl = false;
    int beforeSize, afterSize;
    beforeSize = c.size();
    int i, j;
    CRef confl;

    for (i = 0, j = 0; i < c.size(); i++){
        if (variable_database.value(c[i]) == l_Undef){
            //printf("///@@@ uncheckedEnqueue:index = %d. l_Undef\n", i);
            simpleUncheckEnqueue(~c[i]);
            c[j++] = c[i];
            confl = simplePropagate();
            if (confl != CRef_Undef){
                break;
            }
        }
        else{
            if (variable_database.value(c[i]) == l_True){
                //printf("///@@@ uncheckedEnqueue:index = %d. l_True\n", i);
                c[j++] = c[i];
                True_confl = true;
                confl = implication_graph.reason(var(c[i]));
                break;
            }
            else{
                //printf("///@@@ uncheckedEnqueue:index = %d. l_False\n", i);
                falseLit.push(c[i]);
            }
        }
    }
    c.shrink(c.size() - j);
    afterSize = c.size();
    //printf("\nbefore : %d, after : %d ", beforeSize, afterSize);


    if (confl != CRef_Undef || True_confl == true){
        vec<Lit> simp_learnt_clause;
        simp_learnt_clause.clear();
        vec<CRef> simp_reason_clause;
        simp_reason_clause.clear();
        if (True_confl == true){
            simp_learnt_clause.push(c.last());
        }
        simpleAnalyze(confl, simp_learnt_clause, simp_reason_clause, True_confl);

        if (simp_learnt_clause.size() < c.size()){
            for (i = 0; i < simp_learnt_clause.size(); i++){
                c[i] = simp_learnt_clause[i];
            }
            c.shrink(c.size() - i);
        }
    }

    cancelUntilTrailRecord();
}

bool Solver::simplifyLearnt_core()
{
    int beforeSize, afterSize;
    int learnts_core_size_before = clause_database.learnts_core.size();

    int ci, cj, li, lj;
    bool sat, false_lit;
    unsigned int nblevels;
    ////
    //printf("learnts_x size : %d\n", learnts_x.size());

    ////
    int nbSimplified = 0;
    int nbSimplifing = 0;

    for (ci = 0, cj = 0; ci < clause_database.learnts_core.size(); ci++){
        CRef cr = clause_database.learnts_core[ci];
        Clause& c = clause_database.ca[cr];

        if (removed(cr)) continue;
        else if (c.simplified()){
            clause_database.learnts_core[cj++] = clause_database.learnts_core[ci];
            ////
            nbSimplified++;
        }
        else{
            int saved_size=c.size();
            //         if (drup_file){
            //                 add_oc.clear();
            //                 for (int i = 0; i < c.size(); i++) add_oc.push(c[i]); }
            ////
            nbSimplifing++;
            sat = false_lit = false;
            for (int i = 0; i < c.size(); i++){
                if (variable_database.value(c[i]) == l_True){
                    sat = true;
                    break;
                }
                else if (variable_database.value(c[i]) == l_False){
                    false_lit = true;
                }
            }
            if (sat){
                clause_database.remove_clause_from_database(cr);
            }
            else{
                clause_database.detach_clause_from_watcher_scheme(cr, true);
            	clause_database.update_literal_counters_substract(c);

                if (false_lit){
                    for (li = lj = 0; li < c.size(); li++){
                        if (variable_database.value(c[li]) != l_False){
                            c[lj++] = c[li];
                        }
                    }
                    c.shrink(li - lj);
                }

                beforeSize = c.size();
                assert(c.size() > 1);
                // simplify a learnt clause c
                simplifyLearnt(c);
                assert(c.size() > 0);
                afterSize = c.size();
                
                if(drat_printer.drup_file && saved_size !=c.size()){
#ifdef BIN_DRUP
                    drat_printer.binDRUP('a', c , drat_printer.drup_file);
                    //                    binDRUP('d', add_oc, drup_file);
#else
                    for (int i = 0; i < c.size(); i++)
                        fprintf(drat_printer.drup_file, "%i ", (var(c[i]) + 1) * (-2 * sign(c[i]) + 1));
                    fprintf(drat_printer.drup_file, "0\n");

                    //                    fprintf(drup_file, "d ");
                    //                    for (int i = 0; i < add_oc.size(); i++)
                    //                        fprintf(drup_file, "%i ", (var(add_oc[i]) + 1) * (-2 * sign(add_oc[i]) + 1));
                    //                    fprintf(drup_file, "0\n");
#endif
                }

                //printf("beforeSize: %2d, afterSize: %2d\n", beforeSize, afterSize);

                if (c.size() == 1){
                    // when unit clause occur, enqueue and propagate
                    propagator.uncheckedEnqueue(c[0]);
                    if (propagator.propagate() != CRef_Undef){
                        ok = false;
                        return false;
                    }
                    // delete the clause memory in logic
                    c.mark(1);
                    clause_database.ca.free(cr);
//#ifdef BIN_DRUP
//                    binDRUP('d', c, drup_file);
//#else
//                    fprintf(drup_file, "d ");
//                    for (int i = 0; i < c.size(); i++)
//                        fprintf(drup_file, "%i ", (var(c[i]) + 1) * (-2 * sign(c[i]) + 1));
//                    fprintf(drup_file, "0\n");
//#endif
                }
                else{
                    clause_database.attach_clause_to_watcher_scheme(cr);
                    clause_database.update_literal_counters_add(clause_database.ca[cr]);
                    clause_database.learnts_core[cj++] = clause_database.learnts_core[ci];

                    nblevels = computeLBD(c);
                    if (nblevels < c.lbd()){
                        //printf("lbd-before: %d, lbd-after: %d\n", c.lbd(), nblevels);
                        c.set_lbd(nblevels);
                    }

                    c.setSimplified(true);
                }
            }
        }
    }
    clause_database.learnts_core.shrink(ci - cj);

    //    printf("c nbLearnts_core %d / %d, nbSimplified: %d, nbSimplifing: %d\n",
    //           learnts_core_size_before, learnts_core.size(), nbSimplified, nbSimplifing);

    return true;

}

bool Solver::simplifyLearnt_tier2()
{
    int beforeSize, afterSize;
    int learnts_tier2_size_before = clause_database.learnts_tier2.size();

    int ci, cj, li, lj;
    bool sat, false_lit;
    unsigned int nblevels;
    ////
    //printf("learnts_x size : %d\n", learnts_x.size());

    ////
    int nbSimplified = 0;
    int nbSimplifing = 0;

    for (ci = 0, cj = 0; ci < clause_database.learnts_tier2.size(); ci++){
        CRef cr = clause_database.learnts_tier2[ci];
        Clause& c = clause_database.ca[cr];

        if (removed(cr)) continue;
        else if (c.simplified()){
            clause_database.learnts_tier2[cj++] = clause_database.learnts_tier2[ci];
            ////
            nbSimplified++;
        }
        else{
            int saved_size=c.size();
            //            if (drup_file){
            //                    add_oc.clear();
            //                    for (int i = 0; i < c.size(); i++) add_oc.push(c[i]); }
            ////
            nbSimplifing++;
            sat = false_lit = false;
            for (int i = 0; i < c.size(); i++){
                if (variable_database.value(c[i]) == l_True){
                    sat = true;
                    break;
                }
                else if (variable_database.value(c[i]) == l_False){
                    false_lit = true;
                }
            }
            if (sat){
            	clause_database.remove_clause_from_database(cr);
            }
            else{
                clause_database.detach_clause_from_watcher_scheme(cr, true);
            	clause_database.update_literal_counters_substract(c);

                if (false_lit){
                    for (li = lj = 0; li < c.size(); li++){
                        if (variable_database.value(c[li]) != l_False){
                            c[lj++] = c[li];
                        }
                    }
                    c.shrink(li - lj);
                }

                beforeSize = c.size();
                assert(c.size() > 1);
                // simplify a learnt clause c
                simplifyLearnt(c);
                assert(c.size() > 0);
                afterSize = c.size();
                
                if(drat_printer.drup_file && saved_size!=c.size()){

#ifdef BIN_DRUP
                	drat_printer.binDRUP('a', c , drat_printer.drup_file);
                    //                    binDRUP('d', add_oc, drup_file);
#else
                    for (int i = 0; i < c.size(); i++)
                        fprintf(drat_printer.drup_file, "%i ", (var(c[i]) + 1) * (-2 * sign(c[i]) + 1));
                    fprintf(drat_printer.drup_file, "0\n");

                    //                    fprintf(drup_file, "d ");
                    //                    for (int i = 0; i < add_oc.size(); i++)
                    //                        fprintf(drup_file, "%i ", (var(adddrat_printer.drup_file 1) * (-2 * sign(add_oc[i]) + 1));
                    //                    fprintf(drup_file, "0\n");
#endif
                }

                //printf("beforeSize: %2d, afterSize: %2d\n", beforeSize, afterSize);

                if (c.size() == 1){
                    // when unit clause occur, enqueue and propagate
                    propagator.uncheckedEnqueue(c[0]);
                    if (propagator.propagate() != CRef_Undef){
                        ok = false;
                        return false;
                    }
                    // delete the clause memory in logic
                    c.mark(1);
                    clause_database.ca.free(cr);
//#ifdef BIN_DRUP
//                    binDRUP('d', c, drup_file);
//#else
//                    fprintf(drup_file, "d ");
//                    for (int i = 0; i < c.size(); i++)
//                        fprintf(drup_file, "%i ", (var(c[i]) + 1) * (-2 * sign(c[i]) + 1));
//                    fprintf(drup_file, "0\n");
//#endif
                }
                else{
                    clause_database.attach_clause_to_watcher_scheme(cr);
                    clause_database.update_literal_counters_add(clause_database.ca[cr]);
                    clause_database.learnts_tier2[cj++] = clause_database.learnts_tier2[ci];

                    nblevels = computeLBD(c);
                    if (nblevels < c.lbd()){
                        //printf("lbd-before: %d, lbd-after: %d\n", c.lbd(), nblevels);
                        c.set_lbd(nblevels);
                    }

                    if (c.lbd() <= clause_database.core_lbd_cut){
                        cj--;
                        clause_database.learnts_core.push(cr);
                        c.mark(CORE);
						if (c.watcher_mark() != 0) {
							clause_database.detach_clause_from_watcher_scheme(cr);
							c.watcher_mark(3); // 3 = local
							clause_database.attach_clause_to_watcher_scheme(cr);
						}
                    }

                    c.setSimplified(true);
                }
            }
        }
    }
    clause_database.learnts_tier2.shrink(ci - cj);

    //    printf("c nbLearnts_tier2 %d / %d, nbSimplified: %d, nbSimplifing: %d\n",
    //           learnts_tier2_size_before, learnts_tier2.size(), nbSimplified, nbSimplifing);

    return true;

}

bool Solver::simplifyAll()
{
    if (!ok || propagator.propagate() != CRef_Undef)
        return ok = false;

    if (!simplifyLearnt_core()) return ok = false;
    if (!simplifyLearnt_tier2()) return ok = false;

    checkGarbage();

    return true;
}
//=================================================================================================
// Minor methods:


// Creates a new SAT variable in the solver. If 'decision' is cleared, variable will not be
// used as a decision variable (NOTE! This has effects on the meaning of a SATISFIABLE result).
//
Var Solver::newVar(bool sign, bool dvar)
{
    int v = variable_database.nVars();
    clause_database.watches_bin.init(mkLit(v, false));
    clause_database.watches_bin.init(mkLit(v, true ));
    clause_database.watches  .init(mkLit(v, false));
    clause_database.watches  .init(mkLit(v, true ));
    clause_database.watches_local.init(mkLit(v, false));
    clause_database.watches_local.init(mkLit(v, true ));
    clause_database.watches_high_ranked.init(mkLit(v, false));
    clause_database.watches_high_ranked.init(mkLit(v, true ));

    variable_database.newVar(sign);
    implication_graph.newVar();

    // FIXME: Get these two functions to be one, or even better, get rid of this
    // newVar and have some proper constructors!
    brancher.newVar();
    brancher.setDecisionVar(v, dvar);

    seen     .push(0);
    seen2    .push(0);
    implication_graph.trail    .capacity(v+1);


    pathCs.push(0);
    return v;
}


bool Solver::addClause_(vec<Lit>& ps)
{
    assert(implication_graph.decisionLevel() == 0);
    if (!ok) return false;

    // Check if clause is satisfied and remove false/duplicate literals:
    sort(ps);
    Lit p; int i, j;

    static vec<Lit>            add_oc;
    if (drat_printer.drup_file){
        add_oc.clear();
        for (int i = 0; i < ps.size(); i++) add_oc.push(ps[i]); }

    for (i = j = 0, p = lit_Undef; i < ps.size(); i++)
        if (variable_database.value(ps[i]) == l_True || ps[i] == ~p)
            return true;
        else if (variable_database.value(ps[i]) != l_False && ps[i] != p)
            ps[j++] = p = ps[i];
    ps.shrink(i - j);

    if (drat_printer.drup_file && i != j){
#ifdef BIN_DRUP
    	drat_printer.binDRUP('a', ps, drat_printer.drup_file);
    	drat_printer.binDRUP('d', add_oc, drat_printer.drup_file);
#else
        for (int i = 0; i < ps.size(); i++)
            fprintf(drat_printer.drup_file, "%i ", (var(ps[i]) + 1) * (-2 * sign(ps[i]) + 1));
        fprintf(drat_printer.drup_file, "0\n");

        fprintf(drat_printer.drup_file, "d ");
        for (int i = 0; i < add_oc.size(); i++)
            fprintf(drat_printer.drup_file, "%i ", (var(add_oc[i]) + 1) * (-2 * sign(add_oc[i]) + 1));
        fprintf(drat_printer.drup_file, "0\n");
#endif
    }

    if (ps.size() == 0)
        return ok = false;
    else if (ps.size() == 1){
        propagator.uncheckedEnqueue(ps[0]);
        return ok = (propagator.propagate() == CRef_Undef);
    }else{
        CRef cr = clause_database.ca.alloc(ps, false);
        clause_database.clauses.push(cr);
    	if (clause_database.ca[cr].size() != 2 ) {
    		assert( clause_database.ca[cr].watcher_mark() == 0 ); // 3 = normal
    	} else
    	{
    		clause_database.ca[cr].watcher_mark(2); // 2 = binary
    	}
        clause_database.attach_clause_to_watcher_scheme(cr);
        clause_database.update_literal_counters_add(clause_database.ca[cr]);

    }

    return true;
}

/*
 * TODO: Remove. Is now in VariableDatabase.
 */
//bool Solver::satisfied(const Clause& c) const {
//    for (int i = 0; i < c.size(); i++)
//        if (variable_database.value(c[i]) == l_True)
//            return true;
//    return false; }


// Revert to the state at given level (keeping all assignment at 'level' but not beyond).
//
void Solver::cancelUntil(int bLevel) {
	
    if (implication_graph.decisionLevel() > bLevel){
		static vec<Lit> cancelUntil_tmp;
		cancelUntil_tmp.clear();
        for (int c = implication_graph.trail.size()-1; c >= implication_graph.trail_lim[bLevel]; c--)
        {
            Var      x  = var(implication_graph.trail[c]);

			if (implication_graph.level(x) <= bLevel)
			{
				cancelUntil_tmp.push(implication_graph.trail[c]);
			}
			else
			{
				 brancher.adjust_after_canceling(x);
				 if (phase_saving > 1 || (phase_saving == 1) && c > implication_graph.trail_lim.last())
						variable_database.set_polarity(x,sign(implication_graph.trail[c]));
				variable_database.set_value(x,l_Undef);
			}
        }
        implication_graph.qhead = implication_graph.trail_lim[bLevel];
        implication_graph.trail.shrink(implication_graph.trail.size() - implication_graph.trail_lim[bLevel]);
        implication_graph.trail_lim.shrink(implication_graph.trail_lim.size() - bLevel);
        for (int nLitId = cancelUntil_tmp.size() - 1; nLitId >= 0; --nLitId)
		{
			implication_graph.trail.push_(cancelUntil_tmp[nLitId]);
		}
		
        cancelUntil_tmp.clear();
    } }


//=================================================================================================
// Major methods:


Lit Solver::pickBranchLit()
{
    return brancher.pickBranchLit();
}

inline Solver::ConflictData Solver::FindConflictLevel(CRef cind)
{
	ConflictData data;
	Clause& conflCls = clause_database.ca[cind];
	data.nHighestLevel = implication_graph.level(var(conflCls[0]));
	if (data.nHighestLevel == implication_graph.decisionLevel() && implication_graph.level(var(conflCls[1])) == implication_graph.decisionLevel())
	{
		return data;
	}

	int highestId = 0;
    data.bOnlyOneLitFromHighest = true;
	// find the largest decision level in the clause
	for (int nLitId = 1; nLitId < conflCls.size(); ++nLitId)
	{
		int nLevel = implication_graph.level(var(conflCls[nLitId]));
		if (nLevel > data.nHighestLevel)
		{
			highestId = nLitId;
			data.nHighestLevel = nLevel;
			data.bOnlyOneLitFromHighest = true;
		}
		else if (nLevel == data.nHighestLevel && data.bOnlyOneLitFromHighest == true)
		{
			data.bOnlyOneLitFromHighest = false;
		}
	}

	if (highestId != 0) {
		std::swap(conflCls[0], conflCls[highestId]);
		if (highestId > 1) {
			assert(conflCls.size() != 2 || conflCls.watcher_mark() == 2);
			OccLists<Lit, vec<CleanMaple::ClauseDatabase::Watcher>,
					CleanMaple::ClauseDatabase::WatcherDeleted> &ws =
					conflCls.size() == 2 ? clause_database.watches_bin :
					conflCls.watcher_mark() == 0 ? clause_database.watches :
					conflCls.watcher_mark() == 1 ?
							clause_database.watches_high_ranked :
							clause_database.watches_local;

			//ws.smudge(~conflCls[highestId]);
			remove(ws[~conflCls[highestId]],
					CleanMaple::ClauseDatabase::Watcher(cind, conflCls[1]));
			ws[~conflCls[0]].push(
					CleanMaple::ClauseDatabase::Watcher(cind, conflCls[1]));
		}
	}

	return data;
}


/*_________________________________________________________________________________________________
|
|  analyze : (confl : Clause*) (out_learnt : vec<Lit>&) (out_btlevel : int&)  ->  [void]
|  
|  Description:
|    Analyze conflict and produce a reason clause.
|  
|    Pre-conditions:
|      * 'out_learnt' is assumed to be cleared.
|      * Current decision level must be greater than root level.
|  
|    Post-conditions:
|      * 'out_learnt[0]' is the asserting literal at level 'out_btlevel'.
|      * If out_learnt.size() > 1 then 'out_learnt[1]' has the greatest decision level of the 
|        rest of literals. There may be others from the same level though.
|  
|________________________________________________________________________________________________@*/
void Solver::analyze(CRef confl, vec<Lit>& out_learnt, int& out_btlevel, int& out_lbd, int& out_rank_estimate)
{
	out_learnt.clear();
	out_rank_estimate = 1;

    int pathC = 0;
    Lit p     = lit_Undef;

    // Generate conflict clause:
    //
    out_learnt.push();      // (leave room for the asserting literal)
    int index   = implication_graph.trail.size() - 1;
    int nDecisionLevel = implication_graph.level(var(clause_database.ca[confl][0]));
    assert(nDecisionLevel == implication_graph.level(var(clause_database.ca[confl][0])));

    do{
        assert(confl != CRef_Undef); // (otherwise should be UIP)
        Clause& c = clause_database.ca[confl];

        // For binary clauses, we don't rearrange literals in propagate(), so check and make sure the first is an implied lit.
        if (p != lit_Undef && c.size() == 2 && variable_database.value(c[0]) == l_False){
            assert(variable_database.value(c[1]) == l_True);
            Lit tmp = c[0];
            c[0] = c[1], c[1] = tmp; }

		if (c.watcher_mark() == 0) { // 0 = normal
			clause_database.detach_clause_from_watcher_scheme(confl, true);
			c.watcher_mark(3); // 3 = local
			clause_database.attach_clause_to_watcher_scheme(confl);
		}

		// Update LBD if improved.
		if (c.learnt()) {
			if (out_rank_estimate <= c.rank_estimate()) {
				out_rank_estimate = c.rank_estimate() + 1;
			}
			if (c.mark() != CORE) {
				int lbd = computeLBD(c);
				if (lbd < c.lbd()) {
					if (c.lbd() <= 30)
						c.removable(false); // Protect once from reduction.
					c.set_lbd(lbd);
					if (lbd <= clause_database.core_lbd_cut) {
						clause_database.learnts_core.push(confl);
						c.mark(CORE);
					} else if (lbd <= 6 && c.mark() == LOCAL) {
						// Bug: 'cr' may already be in 'learnts_tier2', e.g., if 'cr' was demoted from TIER2
						// to LOCAL previously and if that 'cr' is not cleaned from 'learnts_tier2' yet.
						clause_database.learnts_tier2.push(confl);
						c.mark(TIER2);
					}
				}

				if (c.mark() == TIER2 || c.mark() == CORE )
					c.touched() = conflicts;
				else if (c.mark() == LOCAL)
					clause_database.claBumpActivity(c);
			}
		}

        for (int j = (p == lit_Undef) ? 0 : 1; j < c.size(); j++){
            Lit q = c[j];
            Var vq = var(q);
            if (!seen[vq] && implication_graph.level(vq) > 0){
                if (brancher.getVsids()){
                	brancher.varBumpActivity(vq, .5);
                	// FIXME: I am not entirely sure, but it seems, that here
                	// add_tmp is solely used by brancher. I.e. we could make
                	// this private member of brancher. See below.
                    add_tmp.push(q);
                }else
                    (brancher.conflicted(vq))++;
                seen[vq] = 1;
                if (implication_graph.level(vq) >= nDecisionLevel){
                    pathC++;
                }else
                    out_learnt.push(q);
            }
        }
        
        // Select next clause to look at:
		do {
			while (!seen[var(implication_graph.trail[index--])]);
			p  = implication_graph.trail[index+1];
		} while (implication_graph.level(var(p)) < nDecisionLevel);
		
        confl = implication_graph.reason(var(p));
        seen[var(p)] = 0;
        pathC--;

    }while (pathC > 0);
    out_learnt[0] = ~p;

    // Simplify conflict clause:
    //
    int i, j;
    out_learnt.copyTo(analyze_toclear);
    if (ccmin_mode == 2){
        uint32_t abstract_level = 0;
        for (i = 1; i < out_learnt.size(); i++)
            abstract_level |= abstractLevel(var(out_learnt[i])); // (maintain an abstraction of levels involved in conflict)

        for (i = j = 1; i < out_learnt.size(); i++)
            if (implication_graph.reason(var(out_learnt[i])) == CRef_Undef || !litRedundant(out_learnt[i], abstract_level))
                out_learnt[j++] = out_learnt[i];
        
    }else if (ccmin_mode == 1){
        for (i = j = 1; i < out_learnt.size(); i++){
            Var x = var(out_learnt[i]);

            if (implication_graph.reason(x) == CRef_Undef)
                out_learnt[j++] = out_learnt[i];
            else{
                Clause& c = clause_database.ca[implication_graph.reason(var(out_learnt[i]))];
                for (int k = c.size() == 2 ? 0 : 1; k < c.size(); k++)
                    if (!seen[var(c[k])] && implication_graph.level(var(c[k])) > 0){
                        out_learnt[j++] = out_learnt[i];
                        break; }
            }
        }
    }else
        i = j = out_learnt.size();

    max_literals += out_learnt.size();
    out_learnt.shrink(i - j);
    tot_literals += out_learnt.size();

    out_lbd = computeLBD(out_learnt);
    if (out_lbd <= 6 && out_learnt.size() <= 30) // Try further minimization?
        if (binResMinimize(out_learnt))
            out_lbd = computeLBD(out_learnt); // Recompute LBD if minimized.

    // Find correct backtrack level:
    //
    if (out_learnt.size() == 1)
        out_btlevel = 0;
    else{
        int max_i = 1;
        // Find the first literal assigned at the next-highest level:
        for (int i = 2; i < out_learnt.size(); i++)
            if (implication_graph.level(var(out_learnt[i])) > implication_graph.level(var(out_learnt[max_i])))
                max_i = i;
        // Swap-in this literal at index 1:
        Lit p             = out_learnt[max_i];
        out_learnt[max_i] = out_learnt[1];
        out_learnt[1]     = p;
        out_btlevel       = implication_graph.level(var(p));
    }

    if (brancher.getVsids()){
        for (int i = 0; i < add_tmp.size(); i++){
            Var v = var(add_tmp[i]);
            if (implication_graph.level(v) >= out_btlevel - 1)
                brancher.varBumpActivity(v, 1);
        }
    	// FIXME: I am not entirely sure, but it seems, that here
    	// add_tmp is solely used by brancher. I.e. we could make
    	// this private member of brancher. See above.
        add_tmp.clear();
    }else{
        seen[var(p)] = true;
        for(int i = out_learnt.size() - 1; i >= 0; i--){
            Var v = var(out_learnt[i]);
            CRef rea = implication_graph.reason(v);
            if (rea != CRef_Undef){
                const Clause& reaC = clause_database.ca[rea];
                for (int i = 0; i < reaC.size(); i++){
                    Lit l = reaC[i];
                    if (!seen[var(l)]){
                        seen[var(l)] = true;
                        (brancher.almost_conflicted(var(l)))++;
                        analyze_toclear.push(l); } } } } }

    for (int j = 0; j < analyze_toclear.size(); j++) seen[var(analyze_toclear[j])] = 0;    // ('seen[]' is now cleared)
}


// Try further learnt clause minimization by means of binary clause resolution.
bool Solver::binResMinimize(vec<Lit>& out_learnt)
{
    // Preparation: remember which false variables we have in 'out_learnt'.
    counter++;
    for (int i = 1; i < out_learnt.size(); i++)
        seen2[var(out_learnt[i])] = counter;

    // Get the list of binary clauses containing 'out_learnt[0]'.
    const vec<CleanMaple::ClauseDatabase::Watcher>& ws = clause_database.watches_bin[~out_learnt[0]];

    int to_remove = 0;
    for (int i = 0; i < ws.size(); i++){
        Lit the_other = ws[i].blocker;
        // Does 'the_other' appear negatively in 'out_learnt'?
        if (seen2[var(the_other)] == counter && variable_database.value(the_other) == l_True){
            to_remove++;
            seen2[var(the_other)] = counter - 1; // Remember to remove this variable.
        }
    }

    // Shrink.
    if (to_remove > 0){
        int last = out_learnt.size() - 1;
        for (int i = 1; i < out_learnt.size() - to_remove; i++)
            if (seen2[var(out_learnt[i])] != counter)
                out_learnt[i--] = out_learnt[last--];
        out_learnt.shrink(to_remove);
    }
    return to_remove != 0;
}


// Check if 'p' can be removed. 'abstract_levels' is used to abort early if the algorithm is
// visiting literals at levels that cannot be removed later.
bool Solver::litRedundant(Lit p, uint32_t abstract_levels)
{
    static vec<Lit>            analyze_stack;
    analyze_stack.clear(); analyze_stack.push(p);
    int top = analyze_toclear.size();
    while (analyze_stack.size() > 0){
        assert(implication_graph.reason(var(analyze_stack.last())) != CRef_Undef);
        Clause& c = clause_database.ca[implication_graph.reason(var(analyze_stack.last()))]; analyze_stack.pop();

        // Special handling for binary clauses like in 'analyze()'.
        if (c.size() == 2 && variable_database.value(c[0]) == l_False){
            assert(variable_database.value(c[1]) == l_True);
            Lit tmp = c[0];
            c[0] = c[1], c[1] = tmp; }

        for (int i = 1; i < c.size(); i++){
            Lit p  = c[i];
            if (!seen[var(p)] && implication_graph.level(var(p)) > 0){
                if (implication_graph.reason(var(p)) != CRef_Undef && (abstractLevel(var(p)) & abstract_levels) != 0){
                    seen[var(p)] = 1;
                    analyze_stack.push(p);
                    analyze_toclear.push(p);
                }else{
                    for (int j = top; j < analyze_toclear.size(); j++)
                        seen[var(analyze_toclear[j])] = 0;
                    analyze_toclear.shrink(analyze_toclear.size() - top);
                    return false;
                }
            }
        }
    }

    return true;
}


/*_________________________________________________________________________________________________
|
|  analyzeFinal : (p : Lit)  ->  [void]
|  
|  Description:
|    Specialized analysis procedure to express the final conflict in terms of assumptions.
|    Calculates the (possibly empty) set of assumptions that led to the assignment of 'p', and
|    stores the result in 'out_conflict'.
|________________________________________________________________________________________________@*/
void Solver::analyzeFinal(Lit p, vec<Lit>& out_conflict)
{
    out_conflict.clear();
    out_conflict.push(p);

    if (implication_graph.decisionLevel() == 0)
        return;

    seen[var(p)] = 1;

    for (int i = implication_graph.trail.size()-1; i >= implication_graph.trail_lim[0]; i--){
        Var x = var(implication_graph.trail[i]);
        if (seen[x]){
            if (implication_graph.reason(x) == CRef_Undef){
                assert(implication_graph.level(x) > 0);
                out_conflict.push(~implication_graph.trail[i]);
            }else{
                Clause& c = clause_database.ca[implication_graph.reason(x)];
                for (int j = c.size() == 2 ? 0 : 1; j < c.size(); j++)
                    if (implication_graph.level(var(c[j])) > 0)
                        seen[var(c[j])] = 1;
            }
            seen[x] = 0;
        }
    }

    seen[var(p)] = 0;
}


/*
 * Returns false in case of toplevel-conflict, true otherwise.
 */
bool Solver::remove_all_satisfied_clauses()
{
    assert(implication_graph.decisionLevel() == 0);

    if (!ok || propagator.propagate() != CRef_Undef)
        return ok = false;

    static int simpDB_assigns = -1;   // Number of top-level assignments since last execution of 'simplify()'.
    if (implication_graph.nAssigns() == simpDB_assigns || (propagator.simpDB_props > 0))
        return true;

    clause_database.removeSatisfied();
    checkGarbage();

    // Why is the following necessary? FIXME
    brancher.rebuildOrderHeap();

    simpDB_assigns = implication_graph.nAssigns();
    /*
     * The following had the comment:
     * 		"(shouldn't depend on stats really, but it will do for now)"
     * Apparently, it needs to be rewritten and the corresponding member
     * function within the clause_database removed.
     */
    propagator.simpDB_props   = clause_database.calculate_simpDB_props();   //

    return true;
}

// pathCs[k] is the number of variables assigned at level k,
// it is initialized to 0 at the begining and reset to 0 after the function execution
bool Solver::collectFirstUIP(CRef confl){
    vec<Lit> involved_lits;
    involved_lits.clear();
    int max_level=1;
    Clause& c=clause_database.ca[confl]; int minLevel=implication_graph.decisionLevel();
    for(int i=0; i<c.size(); i++) {
        Var v=var(c[i]);
        //        assert(!seen[v]);
        if (implication_graph.level(v)>0) {
            seen[v]=1;
            brancher.var_iLevel_tmp(v)=1;
            pathCs[implication_graph.level(v)]++;
            if (minLevel>implication_graph.level(v)) {
                minLevel=implication_graph.level(v);
                assert(minLevel>0);
            }
            //    varBumpActivity(v);
        }
    }

    int limit=implication_graph.trail_lim[minLevel-1];
    for(int i=implication_graph.trail.size()-1; i>=limit; i--) {
        Lit p=implication_graph.trail[i]; Var v=var(p);
        if (seen[v]) {
            int currentDecLevel=implication_graph.level(v);
            //      if (currentDecLevel==decisionLevel())
            //      	varBumpActivity(v);
            seen[v]=0;
            if (--pathCs[currentDecLevel]!=0) {
                Clause& rc=clause_database.ca[implication_graph.reason(v)];
                int reasonVarLevel=(brancher.var_iLevel_tmp(v)+1);
                if(reasonVarLevel>max_level) max_level=reasonVarLevel;
                if (rc.size()==2 && variable_database.value(rc[0])==l_False) {
                    // Special case for binary clauses
                    // The first one has to be SAT
                    assert(variable_database.value(rc[1]) != l_False);
                    Lit tmp = rc[0];
                    rc[0] =  rc[1], rc[1] = tmp;
                }
                for (int j = 1; j < rc.size(); j++){
                    Lit q = rc[j]; Var v1=var(q);
                    if (implication_graph.level(v1) > 0) {
                        if (minLevel>implication_graph.level(v1)) {
                            minLevel=implication_graph.level(v1); limit=implication_graph.trail_lim[minLevel-1]; 	assert(minLevel>0);
                        }
                        if (seen[v1]) {
                            if (brancher.var_iLevel_tmp(v1)<reasonVarLevel)
                                brancher.var_iLevel_tmp(v1)=reasonVarLevel;
                        }
                        else {
                        	brancher.var_iLevel_tmp(v1)=reasonVarLevel;
                            seen[v1] = 1;
                            pathCs[implication_graph.level(v1)]++;
                        }
                    }
                }
            }
            involved_lits.push(p);
        }
    }
    brancher.adjust_while_collecting_first_UIP_outer(involved_lits,max_level );
    return true;
}

struct UIPOrderByILevel_Lt {
    Solver& solver;
    const vec<double>&  var_iLevel;
    bool operator () (Lit x, Lit y) const
    {
        return var_iLevel[var(x)] < var_iLevel[var(y)] ||
                (var_iLevel[var(x)]==var_iLevel[var(y)]&& solver.implication_graph.level(var(x))>solver.implication_graph.level(var(y)));
    }
    UIPOrderByILevel_Lt(const vec<double>&  iLevel, Solver& para_solver) : solver(para_solver), var_iLevel(iLevel) { }
};

void Solver::initialize_search_routine() {
    starts++;

    // Sanity-check:
    assert(ok);

    // Initialization of global variables:
    // Fixme: This is not supposed to be here.
    cached_from_search_method = false;
}

/*
 * Returns l_False in case of conflict, else returns l_True.
 */
lbool Solver::remove_all_satisfied_clauses_within_search() {
    static long curSimplify = 1;
    static int nbconfbeforesimplify = 1000;
    int incSimplify = 1000;
    if (conflicts >= curSimplify * nbconfbeforesimplify){
//        if (!simplifyAll()){
//            return l_False;
//        }
        curSimplify = (conflicts / nbconfbeforesimplify) + 1;
        nbconfbeforesimplify += incSimplify;
    }
    return l_True;
}

void Solver::adjust_brancher_in_handle_conflict_first_time() {
    if (brancher.getVsids()){
    	brancher.adjust_timer_and_var_decay();
    }else
    	brancher.adjust_stepsize();
}


void Solver::adjust_brancher_in_handle_conflict_second_time() {
	// Remark: This function is called after a possible skip of the rest of
	//         the analysis of the conflict due to the fact that the conflict is
	//         only one level above. Hence, the following can not be moved to
	//         the function adjust_brancher_in_handle_conflict_first_time() and
	//         needs to stay here as long as divergence of the behavior to the
	//         original solver is not intended.
    if(conflicts>50000) brancher.setDistance(0);
    else brancher.setDistance(1);
}


void Solver::adjust_Clause_database_in_handle_conflict() {
    if (conflicts == 100000 && clause_database.learnts_core.size() < 100) clause_database.core_lbd_cut = 5;
}

bool Solver::chrono_backtrack_is_desired() {
	return (confl_to_chrono < 0 || confl_to_chrono <= conflicts) && chrono > -1
			&& (implication_graph.decisionLevel() - backtrack_level_from_search_method) >= chrono;
}

void Solver::backtrack_in_search(const ConflictData& data) {
    if (chrono_backtrack_is_desired())
    {
		++chrono_backtrack;
		cancelUntil(data.nHighestLevel -1);
	}
	else // default behavior
	{
		++non_chrono_backtrack;
		cancelUntil(backtrack_level_from_search_method);
	}

}

void Solver::adjust_some_settings_in_search_after_backtrack_before_learning() {
    lbd_from_search_method--;

    if (brancher.getVsids()){
        cached_from_search_method = false;
        conflicts_VSIDS++;
        lbd_queue.push(lbd_from_search_method);
        global_lbd_sum += (lbd_from_search_method > 50 ? 50 : lbd_from_search_method); }
}

void Solver::learn_non_unary_clause_during_search() {
	Minisat::CRef cr = clause_database.learn_non_unary_clause(
			learnt_clause_from_search_method, lbd_from_search_method,
			rank_estimate_from_search_method, conflicts);
	propagator.uncheckedEnqueue(learnt_clause_from_search_method[0],
			backtrack_level_from_search_method, cr);
}

void Solver::print_Drat_during_search() {
    if (drat_printer.drup_file){
#ifdef BIN_DRUP
    	drat_printer.binDRUP('a', learnt_clause_from_search_method, drat_printer.drup_file);
#else
        for (int i = 0; i < learnt_clause_from_search_method.size(); i++)
            fprintf(drat_printer.drup_file, "%i ", (var(learnt_clause_from_search_method[i]) + 1) * (-2 * sign(learnt_clause_from_search_method[i]) + 1));
        fprintf(drat_printer.drup_file, "0\n");
#endif
    }
}

void Solver::learn_clause_during_search() {
    if (learnt_clause_from_search_method.size() == 1){
        propagator.uncheckedEnqueue(learnt_clause_from_search_method[0]);
    }else{
    	learn_non_unary_clause_during_search();
    }
    print_Drat_during_search();
}

void Solver::decay_var_activity() {
    if (brancher.getVsids()) brancher.varDecayActivity();
}

void Solver::decay_activity_during_search() {
	decay_var_activity();
    clause_database.claDecayActivity();
}

/*
 * Returns l_True in case of toplevel-conflict, l_Undef otherwise.
 */
lbool Solver::handle_conflict(CRef& confl) {
	conflicts++;

    adjust_brancher_in_handle_conflict_first_time();
    adjust_Clause_database_in_handle_conflict();

    ConflictData data = FindConflictLevel(confl);
    if (data.nHighestLevel == 0) return l_True;
    if (data.bOnlyOneLitFromHighest)
    {
       	// FIXME: For stuff like this there should be something like a scheduler
        // which takes care of all shortcuts and switches. Like also between
        // VSIDS and LRB and DISTANCE. In particular, these switches should be
        // like starting a new solver.
		cancelUntil(data.nHighestLevel - 1);
		return l_Undef; 	// i.e. skip every analysis and everything else.
							// This is required as long as divergence from the
							// original behavior is not intended.
	}

    adjust_brancher_in_handle_conflict_second_time();

    if(brancher.getVsids() && brancher.getDistance())
        collectFirstUIP(confl);

    analyze(confl, learnt_clause_from_search_method, backtrack_level_from_search_method, lbd_from_search_method, rank_estimate_from_search_method);
    backtrack_in_search(data);
    adjust_some_settings_in_search_after_backtrack_before_learning();
    learn_clause_during_search();
    decay_activity_during_search();
	return l_Undef;
}

bool Solver::restart_in_search_is_desired(const int& nof_conflicts) {
	bool restart = false;
	if (!brancher.getVsids())
		restart = nof_conflicts <= 0;
	else {
		if (!cached_from_search_method) {
			restart =
					lbd_queue.full()
							&& (lbd_queue.avg() * 0.8
									> global_lbd_sum / conflicts_VSIDS);
			cached_from_search_method = true;
		}
	}
	return restart;
}

void Solver::restart_in_search() {
    lbd_queue.clear();
    cached_from_search_method = false;
    // Reached bound on number of conflicts:
    cancelUntil(0);
}






void Solver::reduce_clause_databases_if_desired() {
	static uint64_t next_T2_reduce = 10000;
	if (conflicts >= next_T2_reduce) {
		next_T2_reduce = conflicts + 10000;
		clause_database.reduceDB_Tier2(conflicts);
	}

    static uint64_t next_L_reduce = 15000;
	if (conflicts >= next_L_reduce) {
		next_L_reduce = conflicts + 15000;
		clause_database.reduceDB();

	}

	static uint64_t next_watcher_update = 10000;
	if (conflicts >= next_watcher_update) {
		next_watcher_update = conflicts + 10000;
		clause_database.update_watchers_of_learnts_local();
		clause_database.update_watchers_of_learnts_tier2(conflicts);
		clause_database.update_watchers_of_learnts_core(conflicts);
		clause_database.update_watchers_of_original_clauses(conflicts);
	}

	checkGarbage();
}

/*
 * Returns true if a model is found, false otherwise.
 */
bool Solver::enter_next_branch() {
	Lit next = lit_Undef;

	// New variable decision:
	decisions++;
	next = pickBranchLit();

	if (next == lit_Undef) // Model found!
		return true;

    implication_graph.newDecisionLevel();
    propagator.uncheckedEnqueue(next, implication_graph.decisionLevel());
    return false;
}

/*
 * Returns
 * 	l_True 	if model found,
 * 	l_False if toplevel-conflict is found, and
 * 	l_Undef otherwise.
 */
lbool Solver::handle_no_conflict_in_search()
{
	if (implication_graph.decisionLevel() == 0) {
		// Simplify the set of problem clauses:
		// FIXME: 	Get rid of the negation in the following assignment, which
		//			currently is necessary because I do not want to change the
		//			function remove_all_satisfied_clauses().
		bool toplevel_conflict = !remove_all_satisfied_clauses();
		if (toplevel_conflict)
			return l_False;
	}

	reduce_clause_databases_if_desired();

	bool model_found = enter_next_branch();
	if (model_found)
		return l_True;

	return l_Undef;
}

/*
 * Returns
 * 	l_False if toplevel-conflict is found,
 * 	l_True if model is found,
 * 	l_Undef if restart is performed.
 */
lbool Solver::search(int& nof_conflicts)
{


    // Initialization:
    initialize_search_routine();

    if (remove_all_satisfied_clauses_within_search() == l_False) {
    		return l_False;
    }

    for (;;){
        CRef confl = propagator.propagate();

        if (confl != CRef_Undef){
			// CONFLICT
			nof_conflicts--;
			lbool toplevel_conflict = handle_conflict(confl);
			if (toplevel_conflict == l_True)
				return l_False;
        }else{
            // NO CONFLICT
            if (restart_in_search_is_desired(nof_conflicts)){
            	restart_in_search();
                return l_Undef;
            }
            lbool result = handle_no_conflict_in_search();
            if (result != l_Undef)
            	return result;

        	// FIXME: 	Restart is outside of handle_no_conflict as otherwise,
        	// 			I would need handle_no_conflict to have four possible
        	//			return values, one for each of the following cases:
        	//				1) model found
        	//				2) toplevel conflict found
        	//				3) restart performed
        	//				4) otherwise (in which case the loop is continued,
        	//			where the function search() is supposed to return the
        	//			appropriate lbool in the first tree cases.
        }
    }
}

/*
  Finite subsequences of the Luby-sequence:

  0: 1
  1: 1 1 2
  2: 1 1 2 1 1 2 4
  3: 1 1 2 1 1 2 4 1 1 2 1 1 2 4 8
  ...


 */

static double luby(double y, int x){

    // Find the finite subsequence that contains index 'x', and the
    // size of that subsequence:
    int size, seq;
    for (size = 1, seq = 0; size < x+1; seq++, size = 2*size+1);

    while (size-1 != x){
        size = (size-1)>>1;
        seq--;
        x = x % size;
    }

    return pow(y, seq);
}

static bool switch_mode = false;
static void SIGALRM_switch(int signum) { switch_mode = true; }

void Solver::turn_on_time_dependent_switch_to_VSIDS() {
	if (false) {
		signal(SIGALRM, SIGALRM_switch);
		alarm(2500);
	}
}

void Solver::initialize_solve_routine() {
	turn_on_time_dependent_switch_to_VSIDS();

    model.clear();
    add_tmp.clear();
}

void Solver::initial_VSIDS_phase(lbool& status) {
    brancher.setVsids(true);
    int init = 10000;
    while (status == l_Undef && init > 0 /*&& withinBudget()*/)
        status = search(init);
    brancher.setVsids(false);
}

void Solver::main_search_routine(lbool& status) {
    int curr_restarts = 0;
    while (status == l_Undef /*&& withinBudget()*/){
        if (brancher.getVsids()){
            int weighted = INT32_MAX;
            status = search(weighted);
        }else{
            int nof_conflicts = luby(restart_inc, curr_restarts) * restart_first;
            curr_restarts++;
            status = search(nof_conflicts);
        }
        if (!brancher.getVsids() && switch_mode){
        	brancher.switch_to_VSIDS();
        }
    }
}

void Solver::flush_Drat_in_solve_routine_after_search(const lbool& status) {
    // TODO: Note that the Dratprinter and other utilities need to do everything
    // necessary in such a way, that we can exit the function via _exit()
    // without calling the destructors. However, maybe this is not necessary.
    // TODO: Check the Performance-Penelty of properly calling the destructors.
#ifdef BIN_DRUP
    if (drat_printer.drup_file && status == l_False) drat_printer.binDRUP_flush(drat_printer.drup_file);
#endif
}

void Solver::postprocess_the_result(const lbool& status) {
    if (status == l_True){
        // Extend & copy model:
        model.growTo(variable_database.nVars());
        for (int i = 0; i < variable_database.nVars(); i++) model[i] = variable_database.value(i);
    }else if (status == l_False)
        ok = false;
}

void Solver::cleanup_after_main_search_routine() {
    cancelUntil(0);
}


void Solver::terminate_solve_routine(const lbool& status) {
	flush_Drat_in_solve_routine_after_search(status);
	postprocess_the_result(status);
	cleanup_after_main_search_routine();
}


// NOTE: assumptions passed in member-variable 'assumptions'.
lbool Solver::solve_()
{
	// Sanity-Check:
    if (!ok) return l_False;

    // Initialize:
	initialize_solve_routine();

	// Initialize Return value:
	lbool   status            = l_Undef;

	// The actual search:
    initial_VSIDS_phase(status);
    main_search_routine(status);

    // Terminate:
    terminate_solve_routine(status);

    //  Return:
    return status;
}

//=================================================================================================
// Writing CNF to DIMACS:
// 
// FIXME: this needs to be rewritten completely.

static Var mapVar(Var x, vec<Var>& map, Var& max)
{
    if (map.size() <= x || map[x] == -1){
        map.growTo(x+1, -1);
        map[x] = max++;
    }
    return map[x];
}


void Solver::toDimacs(FILE* f, Clause& c, vec<Var>& map, Var& max)
{
    if (variable_database.satisfied(c)) return;

    for (int i = 0; i < c.size(); i++)
        if (variable_database.value(c[i]) != l_False)
            fprintf(f, "%s%d ", sign(c[i]) ? "-" : "", mapVar(var(c[i]), map, max)+1);
    fprintf(f, "0\n");
}


void Solver::toDimacs(const char *file, const vec<Lit>& assumps)
{
    FILE* f = fopen(file, "wr");
    if (f == NULL)
        fprintf(stderr, "could not open file %s\n", file), exit(1);
    toDimacs(f, assumps);
    fclose(f);
}


void Solver::toDimacs(FILE* f, const vec<Lit>& assumps)
{
    // Handle case when solver is in contradictory state:
    if (!ok){
        fprintf(f, "p cnf 1 2\n1 0\n-1 0\n");
        return; }

    vec<Var> map; Var max = 0;

    // Cannot use removeClauses here because it is not safe
    // to deallocate them at this point. Could be improved.
    int cnt = 0;
    for (int i = 0; i < clause_database.clauses.size(); i++)
        if (!variable_database.satisfied(clause_database.ca[clause_database.clauses[i]]))
            cnt++;

    for (int i = 0; i < clause_database.clauses.size(); i++)
        if (!variable_database.satisfied(clause_database.ca[clause_database.clauses[i]])){
            Clause& c = clause_database.ca[clause_database.clauses[i]];
            for (int j = 0; j < c.size(); j++)
                if (variable_database.value(c[j]) != l_False)
                    mapVar(var(c[j]), map, max);
        }

    fprintf(f, "p cnf %d %d\n", max, cnt);

    for (int i = 0; i < clause_database.clauses.size(); i++)
        toDimacs(f, clause_database.ca[clause_database.clauses[i]], map, max);

    if (verbosity > 0)
        printf("c Wrote %d clauses with %d variables.\n", cnt, max);
}


//=================================================================================================
// Garbage Collection methods:

void Solver::relocAll(ClauseAllocator& to)
{

    clause_database.reallocate(to);
}


void Solver::garbageCollect()
{
	// TODO: Move all of this including the calling function checkGarbage() and
	// similar to the clause_database.


    // Initialize the next region to a size corresponding to the estimated utilization degree. This
    // is not precise but should avoid some unnecessary reallocations for the new region:
    ClauseAllocator to(clause_database.ca.size() - clause_database.ca.wasted());

    relocAll(to);
    if (verbosity >= 2)
        printf("c |  Garbage collection:   %12d bytes => %12d bytes             |\n",
               clause_database.ca.size()*ClauseAllocator::Unit_Size, to.size()*ClauseAllocator::Unit_Size);
    to.moveTo(clause_database.ca);
}

void Solver::printStats() {
	double cpu_time = Minisat::cpuTime();
	double mem_used = Minisat::memUsedPeak();
	printf("c restarts              : %"PRIu64"\n", starts);
	printf("c conflicts             : %-12"PRIu64"   (%.0f /sec)\n",
			conflicts, conflicts / cpu_time);
	printf(
			"c decisions             : %-12"PRIu64"                  (%.0f /sec)\n",
			decisions,
			decisions / cpu_time);
	printf("c propagations          : %-12"PRIu64"   (%.0f /sec)\n",
			propagator.propagations, propagator.propagations / cpu_time);
	printf("c conflict literals     : %-12"PRIu64"   (%4.2f %% deleted)\n",
			tot_literals,
			(max_literals - tot_literals) * 100
					/ (double) max_literals);
	printf(
			"c backtracks            : %-12"PRIu64"   (NCB %0.f%% , CB %0.f%%)\n",
			non_chrono_backtrack + chrono_backtrack,
			(non_chrono_backtrack * 100)
					/ (double) (non_chrono_backtrack
							+ chrono_backtrack),
			(chrono_backtrack * 100)
					/ (double) (non_chrono_backtrack
							+ chrono_backtrack));
	if (mem_used != 0)
		printf("c Memory used           : %.2f MB\n", mem_used);

	printf("c CPU time              : %g s\n", cpu_time);
}
