/****************************************************************************************[Solver.h]
MiniSat -- Copyright (c) 2003-2006, Niklas Een, Niklas Sorensson
           Copyright (c) 2007-2010, Niklas Sorensson

Chanseok Oh's MiniSat Patch Series -- Copyright (c) 2015, Chanseok Oh
 
Maple_LCM, Based on MapleCOMSPS_DRUP -- Copyright (c) 2017, Mao Luo, Chu-Min LI, Fan Xiao: implementing a learnt clause minimisation approach
Reference: M. Luo, C.-M. Li, F. Xiao, F. Manya, and Z. L. , “An effective learnt clause minimization approach for cdcl sat solvers,” in IJCAI-2017, 2017, pp. to–appear.
 
Maple_LCM_Dist, Based on Maple_LCM -- Copyright (c) 2017, Fan Xiao, Chu-Min LI, Mao Luo: using a new branching heuristic called Distance at the beginning of search
 
 
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/

#ifndef Minisat_Solver_h
#define Minisat_Solver_h

#define ANTI_EXPLORATION

#define GLUCOSE23
//#define INT_QUEUE_AVG
//#define LOOSE_PROP_STAT

#ifdef GLUCOSE23
#define INT_QUEUE_AVG
#define LOOSE_PROP_STAT
#endif

#include "../../Utilities/Algorithms/Alg.h"
#include "../../Utilities/BooleanTypes/SolverTypes.h"
#include "../../Utilities/Containers/Vec.h"
#include "../../Utilities/Containers/Heap.h"
#include "../../Utilities/ParseUtilities/OptionsParser/Options.h"
#include "../../Utilities/PrintUtilities/DratPrinter/DratPrinter.h"

#include "Brancher.h"
#include "ClauseDatabase.h"
#include "VariableDatabase.h"
#include "ImplicationGraph.h"
#include "Propagator.h"

// TODO: Get rid of the following if possible. These are also defined in the
// ClauseDatabase.h
// Don't change the actual numbers.
#define LOCAL 0
#define TIER2 2
#define CORE  3

namespace Minisat {

//=================================================================================================
// Solver -- the main class:

class Solver {
private:
    template<typename T>
    class MyQueue {
        int max_sz, q_sz;
        int ptr;
        int64_t sum;
        vec<T> q;
    public:
        MyQueue(int sz) : max_sz(sz), q_sz(0), ptr(0), sum(0) { assert(sz > 0); q.growTo(sz); }
        inline bool   full () const { return q_sz == max_sz; }
#ifdef INT_QUEUE_AVG
        inline T      avg  () const { assert(full()); return sum / max_sz; }
#else
        inline double avg  () const { assert(full()); return sum / (double) max_sz; }
#endif
        inline void   clear()       { sum = 0; q_sz = 0; ptr = 0; }
        void push(T e) {
            if (q_sz < max_sz) q_sz++;
            else sum -= q[ptr];
            sum += e;
            q[ptr++] = e;
            if (ptr == max_sz) ptr = 0;
        }
    };

public:

    // Constructor/Destructor:
    //
    Solver();
    virtual ~Solver();

    // Problem specification:
    //
    Var     newVar    (bool polarity = true, bool dvar = true); // Add a new variable with parameters specifying variable mode.
    bool    addClause_(      vec<Lit>& ps);

    // Solving:
    //
    bool    remove_all_satisfied_clauses     ();                        // Removes already satisfied clauses.
    void    toDimacs     (FILE* f, const vec<Lit>& assumps);            // Write CNF to file in DIMACS-format.
    void    toDimacs     (const char *file, const vec<Lit>& assumps);
    void    toDimacs     (FILE* f, Clause& c, vec<Var>& map, Var& max);
    void    toDimacs     (const char* file);

    // Read state:
    lbool   modelValue (Lit p) const;       // The value of a literal in the last model. The last call to solve must have been satisfiable.
    int     nFreeVars  ()      const;

    // Memory managment:
    virtual void garbageCollect();
    void    checkGarbage(double gf);
    void    checkGarbage();

    // Extra results: (read-only member variable)
    //
    vec<lbool> model;             // If problem is satisfiable, this vector contains the model (if any).
    CleanMaple::DratPrinter drat_printer;


    // Mode of operation:
    //
    int       verbosity;
    int       ccmin_mode;         // Controls conflict clause minimization (0=none, 1=basic, 2=deep).
    int       phase_saving;       // Controls the level of phase saving (0=none, 1=limited, 2=full).

    int       restart_first;      // The initial restart limit.                                                                (default 100)
    double    restart_inc;        // The factor with which the restart limit is multiplied in each restart.                    (default 1.5)

    uint64_t starts, decisions, conflicts, conflicts_VSIDS;
    uint64_t max_literals, tot_literals;
    uint64_t chrono_backtrack, non_chrono_backtrack;




    struct VarOrderLt {
        const vec<double>&  activity;
        bool operator () (Var x, Var y) const { return activity[x] > activity[y]; }
        VarOrderLt(const vec<double>&  act) : activity(act) { }
    };
    
    struct ConflictData
	{
		ConflictData() :
			nHighestLevel(-1),
			bOnlyOneLitFromHighest(false)
		{}

		int nHighestLevel;
		bool bOnlyOneLitFromHighest;
	};

    // Solver state:
    //
    bool                ok;               // If FALSE, the constraints are already unsatisfiable. No part of the solver state may be used!

    CleanMaple::VariableDatabase	variable_database;
    CleanMaple::Brancher 			brancher;
    CleanMaple::ImplicationGraph	implication_graph;

    CleanMaple::ClauseDatabase		clause_database;
    CleanMaple::Propagator			propagator;
    float               global_lbd_sum;
    MyQueue<int>        lbd_queue;  // For computing moving averages of recent LBD values.
    
    int 				confl_to_chrono;
    int 				chrono;

    // Temporaries (to reduce allocation overhead). Each variable is prefixed by the method in which it is
    // used, exept 'seen' wich is used in several places.
    //
    vec<char>           seen;
    vec<Lit>            analyze_toclear;
    vec<Lit>            add_tmp;

    vec<uint64_t>       seen2;    // Mostly for efficient LBD computation. 'seen2[i]' will indicate if decision level or variable 'i' has been seen.
    uint64_t            counter;  // Simple counter for marking purpose with 'seen2'.

    // Main internal methods:
    Lit      pickBranchLit    ();                                                      // Return the next decision variable.
    bool     enqueue          (Lit p, CRef from = CRef_Undef);                         // Test if fact 'p' contradicts current state, enqueue otherwise.

    void     cancelUntil      (int level);                                             // Backtrack until a certain level.
    void     analyze          (CRef confl, vec<Lit>& out_learnt, int& out_btlevel, int& out_lbd);    // (bt = backtrack)
    void     analyzeFinal     (Lit p, vec<Lit>& out_conflict);                         // COULD THIS BE IMPLEMENTED BY THE ORDINARIY "analyze" BY SOME REASONABLE GENERALIZATION?
    bool     litRedundant     (Lit p, uint32_t abstract_levels);                       // (helper method for 'analyze()')

    /////////////////////////////
    // The main solve method, which calles the main search method:
    // TODO: Get all small methods for the search method inlined.
    bool        cached_from_search_method;
    vec<Lit>    learnt_clause_from_search_method;	// for handle conflict. Probably in order to not construct and deconstruct it all the time.
    int         backtrack_level_from_search_method; // for handle conflict. Probably in order to not construct and deconstruct it all the time.
    int			lbd_from_search_method;				// for handle conflict. Probably in order to not construct and deconstruct it all the time.

    void	initialize_search_routine();											  // For the search method.
    lbool	remove_all_satisfied_clauses_within_search();							  // For the search method.
    void	adjust_brancher_in_handle_conflict_first_time();						  // For the search method.
    void	adjust_brancher_in_handle_conflict_second_time();						  // For the search method.
    void	adjust_Clause_database_in_handle_conflict();							  // For the search method.
    bool	chrono_backtrack_is_desired();											  // For the search method.
    void	backtrack_in_search(const ConflictData& data);							  // For the search method.
    void	adjust_some_settings_in_search_after_backtrack_before_learning();		  // For the search method.
    void	learn_non_unary_clause_during_search();									  // For the search method.
    void	print_Drat_during_search();												  // For the search method.
    void	learn_clause_during_search();											  // For the search method.
    void	decay_var_activity();						 							  // For the search method.
    void	decay_activity_during_search();				 							  // For the search method.
    lbool	handle_conflict(CRef& confl);											  // For the search method.
    bool	restart_in_search_is_desired(const int& nof_conflicts);					  // For the search method.
    void	restart_in_search();													  // For the search method.
    void	reduce_clause_databases_if_desired();									  // For the search method.
    bool	enter_next_branch(); 													  // For the search method.
    lbool	handle_no_conflict_in_search();											  // For the search method.
    lbool	search           (int& nof_conflicts);                                    // Search for a given number of conflicts.

/////////////////////////////
// The main solve method, which calls the main search method:

	void	turn_on_time_dependent_switch_to_VSIDS();                                  // For the main solve method.
	void	initialize_solve_routine();                                                // For the main solve method.
    void    initial_VSIDS_phase(lbool& status);                                        // For the main solve method.
    void    main_search_routine(lbool& status);                                        // For the main solve method.
    void    flush_Drat_in_solve_routine_after_search(const lbool& status);             // For the main solve method.
    void    postprocess_the_result(const lbool& status);                               // For the main solve method.
    void    terminate_solve_routine(const lbool& status);                              // For the main solve method.
    void    cleanup_after_main_search_routine();                                       // For the main solve method. .
    lbool	solve_           ();                                                       // Main solve method (assumptions given in 'assumptions').


///////////////////////
    bool     binResMinimize   (vec<Lit>& out_learnt);                                  // Further learnt clause minimization by binary resolution.
    void     relocAll         (ClauseAllocator& to);
    uint32_t abstractLevel    (Var x) const; // Used to represent an abstraction of sets of decision levels.
    ConflictData FindConflictLevel(CRef cind);
    
protected:
    template<class V> int computeLBD(const V& c) {
        int lbd = 0;

        counter++;
        for (int i = 0; i < c.size(); i++){
            int l = implication_graph.level(var(c[i]));
            if (l != 0 && seen2[l] != counter){
                seen2[l] = counter;
                lbd++; } }

        return lbd;
    }

public:
    bool	simplifyAll();
    void	simplifyLearnt(Clause& c);
    bool	simplifyLearnt_x(vec<CRef>& learnts_x);
    bool	simplifyLearnt_core();
    bool	simplifyLearnt_tier2();
    int		trailRecord;
    void	litsEnqueue(int cutP, Clause& c);
    void	cancelUntilTrailRecord();
    void	simpleUncheckEnqueue(Lit p, CRef from = CRef_Undef);
    CRef    simplePropagate();

    void	simpleAnalyze(CRef confl, vec<Lit>& out_learnt, vec<CRef>& reason_clause, bool True_confl);

    // in redundant
    bool removed(CRef cr);
    // adjust simplifyAll occasion

    bool collectFirstUIP(CRef confl);
    vec<int> pathCs;
    void printStats();
};


//=================================================================================================
// Implementation of inline methods:



inline void Solver::checkGarbage(void){ return checkGarbage(clause_database.garbage_frac); }
inline void Solver::checkGarbage(double gf){
    if (clause_database.ca.wasted() > clause_database.ca.size() * gf)
        garbageCollect(); }

// NOTE: enqueue does not set the ok flag! (only public methods do)
inline bool     Solver::enqueue         (Lit p, CRef from)      { return variable_database.value(p) != l_Undef ? variable_database.value(p) != l_False : (propagator.uncheckedEnqueue(p, implication_graph.decisionLevel(), from), true); }

inline uint32_t Solver::abstractLevel (Var x) const   { return 1 << (implication_graph.level(x) & 31); }
inline lbool    Solver::modelValue    (Lit p) const   { return model[var(p)] ^ sign(p); }
inline int      Solver::nFreeVars     ()      const   { return (int)brancher.get_dec_vars() - (implication_graph.trail_lim.size() == 0 ? implication_graph.trail.size() : implication_graph.trail_lim[0]); }

inline void     Solver::toDimacs     (const char* file){ vec<Lit> as; toDimacs(file, as); }



//=================================================================================================
// Debug etc:


//=================================================================================================
}

#endif
