/***********************************************************************************[SimpSolver.cc]
MiniSat -- Copyright (c) 2006,      Niklas Een, Niklas Sorensson
           Copyright (c) 2007-2010, Niklas Sorensson

Chanseok Oh's MiniSat Patch Series -- Copyright (c) 2015, Chanseok Oh
 
Maple_LCM, Based on MapleCOMSPS_DRUP -- Copyright (c) 2017, Mao Luo, Chu-Min LI, Fan Xiao: implementing a learnt clause minimisation approach
Reference: M. Luo, C.-M. Li, F. Xiao, F. Manya, and Z. L. , “An effective learnt clause minimization approach for cdcl sat solvers,” in IJCAI-2017, 2017, pp. to–appear.
 
Maple_LCM_Dist, Based on Maple_LCM -- Copyright (c) 2017, Fan Xiao, Chu-Min LI, Mao Luo: using a new branching heuristic called Distance at the beginning of search
 
 
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/

#include "SimpSolver.h"

#include "../../Utilities/Algorithms/Sort.h"
#include "../../Utilities/System/System.h"

using namespace Minisat;

//=================================================================================================
// Options:


static const char* _cat = "SIMP";

static BoolOption   opt_use_elim         (_cat, "elim",         "Perform variable elimination.", true);
static IntOption    opt_grow             (_cat, "grow",         "Allow a variable elimination step to grow by a number of clauses.", 0);
static IntOption    opt_clause_lim       (_cat, "cl-lim",       "Variables are not eliminated if it produces a resolvent with a length above this limit. -1 means no limit", 20,   IntRange(-1, INT32_MAX));
static IntOption    opt_subsumption_lim  (_cat, "sub-lim",      "Do not check if subsumption against a clause larger than this. -1 means no limit.", 1000, IntRange(-1, INT32_MAX));
static DoubleOption opt_simp_garbage_frac(_cat, "simp-gc-frac", "The fraction of wasted memory allowed before a garbage collection is triggered during simplification.",  0.5, DoubleRange(0, false, HUGE_VAL, false));


//=================================================================================================
// Constructor/Destructor:


SimpSolver::SimpSolver() :
    parsing            (false)
  , grow               (opt_grow)
  , clause_lim         (opt_clause_lim)
  , subsumption_lim    (opt_subsumption_lim)
  , simp_garbage_frac  (opt_simp_garbage_frac)
  , use_elim           (opt_use_elim)
  , merges             (0)
  , eliminated_vars    (0)
  , elimorder          (1)
  , use_simplification (true)
  , occurs             (ClauseDeleted(clause_database.ca))
  , elim_heap          (ElimLt(n_occ))
  , bwdsub_assigns     (0)
  , n_touched          (0)
{
    vec<Lit> dummy(1,lit_Undef);
    clause_database.ca.extra_clause_field = true; // NOTE: must happen before allocating the dummy clause below.
    bwdsub_tmpunit        = clause_database.ca.alloc(dummy);
    clause_database.remove_satisfied      = false;
}


SimpSolver::~SimpSolver()
{
}



Var SimpSolver::newVar(bool sign, bool dvar) {
    Var v = Solver::newVar(sign, dvar);

    frozen    .push((char)false);
    eliminated.push((char)false);

    if (use_simplification){
        n_occ     .push(0);
        n_occ     .push(0);
        occurs    .init(v);
        touched   .push(0);
        elim_heap .insert(v);
    }
    return v; }



bool SimpSolver::addClause_(vec<Lit>& ps)
{
#ifndef NDEBUG
    for (int i = 0; i < ps.size(); i++)
        assert(!isEliminated(var(ps[i])));
#endif

    int nclauses = clause_database.clauses.size();

    if (!Solver::addClause_(ps))
        return false;

    if (!parsing && drat_printer.drup_file) {
#ifdef BIN_DRUP
    	drat_printer.binDRUP('a', ps, drat_printer.drup_file);
#else
        for (int i = 0; i < ps.size(); i++)
            fprintf(drat_printer.drup_file, "%i ", (var(ps[i]) + 1) * (-2 * sign(ps[i]) + 1));
        fprintf(drat_printer.drup_file, "0\n");
#endif
    }

    if (use_simplification && clause_database.clauses.size() == nclauses + 1){
        CRef          cr = clause_database.clauses.last();
        const Clause& c  = clause_database.ca[cr];

        // NOTE: the clause is added to the queue immediately and then
        // again during 'gatherTouchedClauses()'. If nothing happens
        // in between, it will only be checked once. Otherwise, it may
        // be checked twice unnecessarily. This is an unfortunate
        // consequence of how backward subsumption is used to mimic
        // forward subsumption.
        subsumption_queue.insert(cr);
        for (int i = 0; i < c.size(); i++){
            occurs[var(c[i])].push(cr);
            n_occ[toInt(c[i])]++;
            touched[var(c[i])] = 1;
            n_touched++;
            if (elim_heap.inHeap(var(c[i])))
                elim_heap.increase(var(c[i]));
        }
    }

    return true;
}


void SimpSolver::removeClause(CRef cr)
{
    const Clause& c = clause_database.ca[cr];

    if (use_simplification)
        for (int i = 0; i < c.size(); i++){
            n_occ[toInt(c[i])]--;
            updateElimHeap(var(c[i]));
            occurs.smudge(var(c[i]));
        }

	clause_database.remove_clause_from_database(cr);
}


bool SimpSolver::strengthenClause(CRef cr, Lit l)
{
    Clause& c = clause_database.ca[cr];
    assert(implication_graph.decisionLevel() == 0);
    assert(use_simplification);

    // FIX: this is too inefficient but would be nice to have (properly implemented)
    // if (!find(subsumption_queue, &c))
    subsumption_queue.insert(cr);

    if (drat_printer.drup_file){
#ifdef BIN_DRUP
    	drat_printer.binDRUP_strengthen(c, l, drat_printer.drup_file);
#else
        for (int i = 0; i < c.size(); i++)
            if (c[i] != l) fprintf(drat_printer.drup_file, "%i ", (var(c[i]) + 1) * (-2 * sign(c[i]) + 1));
        fprintf(drat_printer.drup_file, "0\n");
#endif
    }

    if (c.size() == 2){
        removeClause(cr);
        c.strengthen(l);
    }else{
        if (drat_printer.drup_file){
#ifdef BIN_DRUP
        	drat_printer.binDRUP('d', c, drat_printer.drup_file);
#else
            fprintf(drat_printer.drup_file, "d ");
            for (int i = 0; i < c.size(); i++)
                fprintf(drat_printer.drup_file, "%i ", (var(c[i]) + 1) * (-2 * sign(c[i]) + 1));
            fprintf(drat_printer.drup_file, "0\n");
#endif
        }

        clause_database.detach_clause_from_watcher_scheme(cr, true);
    	clause_database.update_literal_counters_substract(c);
        c.strengthen(l);
        if (c.size() == 2 ) {
    		c.watcher_mark(2); // 2 = binary
    	}
        clause_database.attach_clause_to_watcher_scheme(cr);
        clause_database.update_literal_counters_add(clause_database.ca[cr]);

        remove(occurs[var(l)], cr);
        n_occ[toInt(l)]--;
        updateElimHeap(var(l));
    }

    return c.size() == 1 ? enqueue(c[0]) && propagator.propagate() == CRef_Undef : true;
}


// Returns FALSE if clause is always satisfied ('out_clause' should not be used).
bool SimpSolver::compute_resolvent(const Clause& _ps, const Clause& _qs, Var v, vec<Lit>& out_clause)
{
	// FIXME: Rewrite this function to be better readable. Its horrible.

	/*
	 * The chosen algorithm has o(size(_ps)*size(_qs)) complexity. There is an
	 * algorithm with o(size(_ps)+size(_qs)) complexity, that may be chosen
	 * instead for large size(_ps)*size(_qs) counting seen literals.
	 * However, mostly this is small counting seen literals is probably slower,
	 * in these cases. But this is not tested. So TODO: Check other algorithm.
	 */

	/*
	 * Also one might take advantage of the fact that the function
	 * count_literals_in_resolvent was called before. Hence, essentially, this
	 * clause was computed. Maybe, we might be able to skip some computations at
	 * least in the cases where the resolvent is always satisfied.
	 */

    merges++;
    out_clause.clear();

    bool  ps_smallest = _ps.size() < _qs.size();
    const Clause& ps  =  ps_smallest ? _qs : _ps;
    const Clause& qs  =  ps_smallest ? _ps : _qs;

    for (int i = 0; i < qs.size(); i++){
        if (var(qs[i]) != v){
            for (int j = 0; j < ps.size(); j++)
                if (var(ps[j]) == var(qs[i]))
                    if (ps[j] == ~qs[i])
                        return false;
                    else
                        goto next;
            out_clause.push(qs[i]);
        }
        next:;
    }

    for (int i = 0; i < ps.size(); i++)
        if (var(ps[i]) != v)
            out_clause.push(ps[i]);

    return true;
}


// Returns FALSE if resolvent is always satisfied.
bool SimpSolver::count_literals_in_resolvent(const Clause& _ps, const Clause& _qs, Var v, int& size)
{
	// FIXME: Rewrite this function to be better readable. Its horrible.

	/*
	 * The chosen algorithm has o(size(_ps)*size(_qs)) complexity. There is an
	 * algorithm with o(size(_ps)+size(_qs)) complexity, that may be chosen
	 * instead for large size(_ps)*size(_qs) counting seen literals.
	 * However, mostly this is small counting seen literals is probably slower,
	 * in these cases. But this is not tested. So TODO: Check other algorithm.
	 */

    merges++;

    bool  ps_smallest = _ps.size() < _qs.size();
    const Clause& ps  =  ps_smallest ? _qs : _ps;
    const Clause& qs  =  ps_smallest ? _ps : _qs;
    const Lit*  __ps  = (const Lit*)ps;
    const Lit*  __qs  = (const Lit*)qs;

    size = ps.size()-1;

    for (int i = 0; i < qs.size(); i++){
        if (var(__qs[i]) != v){
            for (int j = 0; j < ps.size(); j++)
                if (var(__ps[j]) == var(__qs[i]))
                    if (__ps[j] == ~__qs[i])
                        return false;
                    else
                        goto next;
            size++;
        }
        next:;
    }

    return true;
}


void SimpSolver::gatherTouchedClauses()
{
    if (n_touched == 0) return;

    /*
     * Initially all clauses are constructed with mark = 0. (At least the non-
     * learnts important here.) Possibly, the mark cannot be set to be 1,
     * because 1 means deleted and that might interfere with other stuff, I
     * don't know. The choice to be 2 seems random. FIXME: Do better.
     */
    int i,j;
    for (i = j = 0; i < subsumption_queue.size(); i++)
        if (clause_database.ca[subsumption_queue[i]].mark() == 0)
            clause_database.ca[subsumption_queue[i]].mark(2);

    for (i = 0; i < touched.size(); i++)
        if (touched[i]){
            const vec<CRef>& cs = occurs.lookup(i);
            for (j = 0; j < cs.size(); j++)
                if (clause_database.ca[cs[j]].mark() == 0){
                    subsumption_queue.insert(cs[j]);
                    clause_database.ca[cs[j]].mark(2);
                }
            touched[i] = 0;
        }

    for (i = 0; i < subsumption_queue.size(); i++)
        if (clause_database.ca[subsumption_queue[i]].mark() == 2)
            clause_database.ca[subsumption_queue[i]].mark(0);

    n_touched = 0;
}

Var SimpSolver::var_with_least_occurences(Clause &c) {
	// Find best variable to scan:
	Var best = var(c[0]);
	for (int i = 1; i < c.size(); i++)
		if (occurs[var(c[i])].size() < occurs[best].size())
			best = var(c[i]);
	return best;
}

// Backward subsumption + backward subsumption resolution
bool SimpSolver::backwardSubsumptionCheck(bool verbose)
{
    int cnt = 0;
    int subsumed = 0;
    int deleted_literals = 0;
    assert(implication_graph.decisionLevel() == 0);

    while (subsumption_queue.size() > 0 || bwdsub_assigns < implication_graph.trail.size()){

    	// FIXME: What is happening here?
        // Check top-level assignments by creating a dummy clause and placing it in the queue:
		if (subsumption_queue.size() == 0 && bwdsub_assigns < implication_graph.trail.size()) {
			Lit l = implication_graph.trail[bwdsub_assigns++];
			clause_database.ca[bwdsub_tmpunit][0] = l;
			clause_database.ca[bwdsub_tmpunit].calcAbstraction();
			subsumption_queue.insert(bwdsub_tmpunit);
		}

        CRef    cr = subsumption_queue.peek(); subsumption_queue.pop();
        Clause& c  = clause_database.ca[cr];

        if (c.mark()) continue;

        if (verbose && verbosity >= 2 && cnt++ % 1000 == 0)
            printf("c subsumption left: %10d (%10d subsumed, %10d deleted literals)\r", subsumption_queue.size(), subsumed, deleted_literals);

        assert(c.size() > 1 || variable_database.value(c[0]) == l_True);    // Unit-clauses should have been propagated before this point.

		Var v = var_with_least_occurences(c);

        // Search all candidates:
        vec<CRef>& _cs = occurs.lookup(v);
        CRef*       cs = (CRef*)_cs;

        for (int j = 0; j < _cs.size(); j++)
            if (c.mark()) // i.e. has been deleted
                break;
            else if (!clause_database.ca[cs[j]].mark() &&  cs[j] != cr && (subsumption_lim == -1 || clause_database.ca[cs[j]].size() < subsumption_lim)){
                Lit l = c.subsumes(clause_database.ca[cs[j]]);

                if (l == lit_Undef)
                    subsumed++, removeClause(cs[j]);
                else if (l != lit_Error){
                    deleted_literals++;

                    if (!strengthenClause(cs[j], ~l))
                        return false;

                    // Did current candidate get deleted from cs? Then check candidate at index j again:
                    if (var(l) == v)
                        j--;
                }
            }
    }

    return true;
}



static void mkElimClause(vec<uint32_t>& elimclauses, Lit x)
{
    elimclauses.push(toInt(x));
    elimclauses.push(1);
}


static void mkElimClause(vec<uint32_t>& elimclauses, Var v, Clause& c)
{
    int first = elimclauses.size();
    int v_pos = -1;

    // Copy clause to elimclauses-vector. Remember position where the
    // variable 'v' occurs:
    for (int i = 0; i < c.size(); i++){
        elimclauses.push(toInt(c[i]));
        if (var(c[i]) == v)
            v_pos = i + first;
    }
    assert(v_pos != -1);

    // Swap the first literal with the 'v' literal, so that the literal
    // containing 'v' will occur first in the clause:
    uint32_t tmp = elimclauses[v_pos];
    elimclauses[v_pos] = elimclauses[first];
    elimclauses[first] = tmp;

    // Store the length of the clause last:
    elimclauses.push(c.size());
}

/*
 * Checks whether the increase in number of clauses stays within the allowed
 * ('grow') and moreover, whether no resolvent exceeds the limit on the maximal
 * clause size (if it is set).
 *
 *
 */
bool SimpSolver::elimination_is_desired(const vec<CRef> &pos,
		const vec<CRef> &neg, Var v, const vec<CRef> &cls) {
	int cnt = 0;
	int clause_size = 0;
	for (int i = 0; i < pos.size(); i++) {
		for (int j = 0; j < neg.size(); j++) {
			bool resolvent_is_not_yet_satisfied = count_literals_in_resolvent(
					clause_database.ca[pos[i]], clause_database.ca[neg[j]], v, clause_size);
			if (resolvent_is_not_yet_satisfied) {
				if (++cnt > cls.size() + grow
						|| (clause_lim != -1 && clause_size > clause_lim)) {
					return false;
				}
			}
		}
	}
	return true;
}

/*
 * Checks whether the increase in number of clauses stays within the allowed 
 * ('grow') and moreover, whether the clause exceeds the limit on the maximal 
 * clause size (if it is set).
 * 
 * Expects cls to be the occurrences of v, pos the positive occurrences and neg 
 * the negative occurrences. These are calculated earlier, so this speeds it up
 * a little. However, the effect of this speedup, I did not check. 
 */
bool SimpSolver::elimination_is_desired(Var& v, const vec<CRef>& cls, vec<CRef>& pos,vec<CRef>& neg) {

	int cnt = 0;
	int clause_size = 0;

	for (int i = 0; i < pos.size(); i++) {
		for (int j = 0; j < neg.size(); j++) {
			bool resolvent_is_not_yet_satisfied = count_literals_in_resolvent(
					clause_database.ca[pos[i]], clause_database.ca[neg[j]], v, clause_size);
			if (resolvent_is_not_yet_satisfied
					&& (++cnt > cls.size() + grow
							|| (clause_lim != -1 && clause_size > clause_lim)))
			{
				return false;
			}
		}
	}
	return true;
}

Var SimpSolver::delete_and_store_to_be_resolved_clauses(Var v,
		const vec<CRef> &pos, const vec<CRef> &neg) {
	// Delete and store old clauses:
	eliminated[v] = true;
	brancher.setDecisionVar(v, false);
	eliminated_vars++;
	if (pos.size() > neg.size()) {
		for (int i = 0; i < neg.size(); i++)
			mkElimClause(elimclauses, v, clause_database.ca[neg[i]]);
		mkElimClause(elimclauses, mkLit(v));
	} else {
		for (int i = 0; i < pos.size(); i++)
			mkElimClause(elimclauses, v, clause_database.ca[pos[i]]);
		mkElimClause(elimclauses, ~mkLit(v));
	}
	return v;
}

void SimpSolver::compute_resolvents(const vec<CRef> &pos, const vec<CRef> &neg,
		Var v) {
	// Produce clauses in cross product:
	vec<Lit> &resolvent = add_tmp;
	for (int i = 0; i < pos.size(); i++) {
		for (int j = 0; j < neg.size(); j++) {
			if (compute_resolvent(clause_database.ca[pos[i]], clause_database.ca[neg[j]], v, resolvent)
					&& !addClause_(resolvent)) {
				// return false;
			}
		}
	}
}

void SimpSolver::cleanup_after_Variable_elimination(const vec<CRef> &cls,
		Var v) {
	for (int i = 0; i < cls.size(); i++)
		removeClause(cls[i]);
	// Free occurs list for this variable:
	occurs[v].clear(true);
	// Free watchers lists for this variable, if possible:
	clause_database.watches_bin[mkLit(v)].clear(true);
	clause_database.watches_bin[~mkLit(v)].clear(true);
	clause_database.watches[mkLit(v)].clear(true);
	clause_database.watches[~mkLit(v)].clear(true);
	clause_database.watches_local[mkLit(v)].clear(true);
	clause_database.watches_local[~mkLit(v)].clear(true);
	clause_database.watches_high_ranked[mkLit(v)].clear(true);
	clause_database.watches_high_ranked[~mkLit(v)].clear(true);

}

/*
 * Returns false if toplevel-conflict is found. Otherwise, returns true.
 */
bool SimpSolver::eliminateVar(Var v)
{

	assert(!frozen[v]);
    assert(!isEliminated(v));
    assert(variable_database.value(v) == l_Undef);

	// Split the occurrences into positive and negative:
	//
	const vec<CRef> &cls = occurs.lookup(v);
	vec<CRef> pos, neg;
	for (int i = 0; i < cls.size(); i++)
		(find(clause_database.ca[cls[i]], mkLit(v)) ? pos : neg).push(cls[i]);

	if (elimination_is_desired(v, cls, pos, neg)) {
		v = delete_and_store_to_be_resolved_clauses(v, pos, neg);
		compute_resolvents(pos, neg, v);
		cleanup_after_Variable_elimination(cls, v);
		return backwardSubsumptionCheck();
	} else {
		return true;
	}
}





void SimpSolver::extendModel()
{
    int i, j;
    Lit x;

	for (i = elimclauses.size() - 1; i > 0; i -= j) {
		for (j = elimclauses[i--]; j > 1; j--, i--)
			if (modelValue(toLit(elimclauses[i])) != l_False)
				goto next;

		x = toLit(elimclauses[i]);
		model[var(x)] = lbool(!sign(x));
		next: ;
	}
}

// Almost duplicate of Solver::removeSatisfied. Didn't want to make the base method 'virtual'.
void SimpSolver::removeSatisfied()
{
    int i, j;
    for (i = j = 0; i < clause_database.clauses.size(); i++){
        const Clause& c = clause_database.ca[clause_database.clauses[i]];
        if (c.mark() == 0)
            if (variable_database.satisfied(c))
                removeClause(clause_database.clauses[i]);
            else
            	clause_database.clauses[j++] = clause_database.clauses[i];
    }
    clause_database.clauses.shrink(i - j);
}

void SimpSolver::initialize_if_no_pre() {
	cleanup_after_eliminate();
}

void SimpSolver::cleanup_after_eliminate() {
	touched.clear(true);
	occurs.clear(true);
	n_occ.clear(true);
	elim_heap.clear(true);
	subsumption_queue.clear(true);
	use_simplification = false;
	clause_database.remove_satisfied = true;
	//clause_database.ca.extra_clause_field = false;
	// Force full cleanup (this is safe and desirable since it only happens once):
	brancher.rebuildOrderHeap();
	garbageCollect();
}

// The technique and code are by the courtesy of the GlueMiniSat team. Thank you!
// It helps solving certain types of huge problems tremendously.
bool SimpSolver::eliminate(bool turn_off_elim)
{
    bool result = true;
    int iter = 0;
    int n_cls, n_cls_init, n_vars;

    //FIXME: What if?
    assert(variable_database.nVars() != 0);

    // Get an initial number of clauses (more accurately).
    if (implication_graph.trail.size() != 0) removeSatisfied();
    n_cls_init = clause_database.clauses.size();

    result = eliminate_(); // The first, usual variable elimination of MiniSat.
    if (!result) {
    	cleanup_after_eliminate();
    	return result;
    }

    n_cls  = clause_database.clauses.size();
    n_vars = nFreeVars();

    printf("c Reduced to %d vars, %d cls (grow=%d)\n", n_vars, n_cls, grow);

    if ((double)n_cls / n_vars >= 10 || n_vars < 10000){
        printf("c No iterative elimination performed. (vars=%d, c/v ratio=%.1f)\n", n_vars, (double)n_cls / n_vars);
        cleanup_after_eliminate();
        return result;
    }

    grow = grow ? grow * 2 : 8;
    for (; grow < 10000; grow *= 2){
        // Rebuild elimination variable heap.
        for (int i = 0; i < clause_database.clauses.size(); i++){
            const Clause& c = clause_database.ca[clause_database.clauses[i]];
            for (int j = 0; j < c.size(); j++)
                if (!elim_heap.inHeap(var(c[j])))
                    elim_heap.insert(var(c[j]));
                else
                    elim_heap.update(var(c[j])); }

        int n_cls_last = clause_database.clauses.size();
        int n_vars_last = nFreeVars();

        result = eliminate_();
        if (!result || n_vars_last == nFreeVars()) break;
        iter++;

        int n_cls_now  = clause_database.clauses.size();
        int n_vars_now = nFreeVars();

        double cl_inc_rate  = (double)n_cls_now   / n_cls_last;
        double var_dec_rate = (double)n_vars_last / n_vars_now;

        printf("c Reduced to %d vars, %d cls (grow=%d)\n", n_vars_now, n_cls_now, grow);
        printf("c cl_inc_rate=%.3f, var_dec_rate=%.3f\n", cl_inc_rate, var_dec_rate);

        if (n_cls_now > n_cls_init || cl_inc_rate > var_dec_rate) break;
    }
    printf("c No. effective iterative eliminations: %d\n", iter);

    cleanup_after_eliminate();
    return result;
}

void SimpSolver::cleanup_after_eliminate_(int trail_size_last) {
	// To get an accurate number of clauses.
	if (trail_size_last != implication_graph.trail.size())
		removeSatisfied();
	else {
		int i, j;
		for (i = j = 0; i < clause_database.clauses.size(); i++)
			if (clause_database.ca[clause_database.clauses[i]].mark() == 0)
				clause_database.clauses[j++] = clause_database.clauses[i];
		clause_database.clauses.shrink(i - j);
	}
	checkGarbage();
	if (verbosity >= 1 && elimclauses.size() > 0)
		printf(
				"c |  Eliminated clauses:     %10.2f Mb                                      |\n",
				double(elimclauses.size() * sizeof(uint32_t)) / (1024 * 1024));
}

bool SimpSolver::can_be_eliminated_(Var elim) {
	return !isEliminated(elim) && variable_database.value(elim) == l_Undef && !frozen[elim];
}

bool SimpSolver::subsumption_check_is_desired() {
	return subsumption_queue.size() > 0 || bwdsub_assigns < implication_graph.trail.size();
}

bool SimpSolver::eliminate_()
{
    if (!remove_all_satisfied_clauses())
        return false;
    else if (!use_simplification)
        return true;

    int trail_size_last = implication_graph.trail.size();

    while (n_touched > 0 || bwdsub_assigns < implication_graph.trail.size() || elim_heap.size() > 0){

		gatherTouchedClauses();
		if (subsumption_check_is_desired()) {
			if (!backwardSubsumptionCheck(true)) {
				ok = false;
				cleanup_after_eliminate_(trail_size_last);
				return ok;
			}
		}

        for (int cnt = 0; !elim_heap.empty(); cnt++){
			Var elim = elim_heap.removeMin();

			if (can_be_eliminated_(elim)) {
				if (!eliminateVar(elim)) {
					ok = false;
					cleanup_after_eliminate_(trail_size_last);
					return ok;
				}
			}

            checkGarbage(simp_garbage_frac);
        }

        assert(subsumption_queue.size() == 0);
    }

    cleanup_after_eliminate_(trail_size_last);
    return ok;
}


//=================================================================================================
// Garbage Collection methods:


void SimpSolver::relocAll(ClauseAllocator& to)
{
    if (!use_simplification) return;

    // All occurs lists:
    //
    occurs.cleanAll();
    for (int i = 0; i < variable_database.nVars(); i++){
        vec<CRef>& cs = occurs[i];
        for (int j = 0; j < cs.size(); j++)
            clause_database.ca.reloc(cs[j], to);
    }

    // Subsumption queue:
    //
    for (int i = 0; i < subsumption_queue.size(); i++)
        clause_database.ca.reloc(subsumption_queue[i], to);

    // Temporary clause:
    //
    clause_database.ca.reloc(bwdsub_tmpunit, to);
}


void SimpSolver::garbageCollect()
{
    // Initialize the next region to a size corresponding to the estimated utilization degree. This
    // is not precise but should avoid some unnecessary reallocations for the new region:
    ClauseAllocator to(clause_database.ca.size() - clause_database.ca.wasted());

    to.extra_clause_field = clause_database.ca.extra_clause_field; // NOTE: this is important to keep (or lose) the extra fields.
    relocAll(to);
    Solver::relocAll(to);
    if (verbosity >= 2)
        printf("c |  Garbage collection:   %12d bytes => %12d bytes             |\n", 
               clause_database.ca.size()*ClauseAllocator::Unit_Size, to.size()*ClauseAllocator::Unit_Size);
    to.moveTo(clause_database.ca);
}
