/************************************************************************************[SimpSolver.h]
MiniSat -- Copyright (c) 2006,      Niklas Een, Niklas Sorensson
           Copyright (c) 2007-2010, Niklas Sorensson

Chanseok Oh's MiniSat Patch Series -- Copyright (c) 2015, Chanseok Oh
 
Maple_LCM, Based on MapleCOMSPS_DRUP -- Copyright (c) 2017, Mao Luo, Chu-Min LI, Fan Xiao: implementing a learnt clause minimisation approach
Reference: M. Luo, C.-M. Li, F. Xiao, F. Manya, and Z. L. , “An effective learnt clause minimization approach for cdcl sat solvers,” in IJCAI-2017, 2017, pp. to–appear.

Maple_LCM_Dist, Based on Maple_LCM -- Copyright (c) 2017, Fan Xiao, Chu-Min LI, Mao Luo: using a new branching heuristic called Distance at the beginning of search
 
 
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/

#ifndef Minisat_SimpSolver_h
#define Minisat_SimpSolver_h

#include "../CoreSolver/Solver.h"
#include "../../Utilities/Containers/Queue.h"

namespace Minisat {

//=================================================================================================


class SimpSolver : public Solver {
 public:
    // Constructor/Destructor:
    //
    SimpSolver();
    ~SimpSolver();

    // Problem specification:
    Var 	newVar(bool sign = true, bool dvar = true);
    bool    addClause_(      vec<Lit>& ps);

    // Variable mode:
    bool    isEliminated(Var v) const;

    // Solving:
    bool    eliminate   (bool turn_off_elim = false);  // Perform variable elimination based simplification. 
    bool    eliminate_  ();
    void    removeSatisfied();

    // Memory management:
    virtual void garbageCollect();

	// Mode of operation:
	bool parsing;
	int grow; // Allow a variable elimination step to grow by a number of clauses (default to zero).
	int clause_lim; // Variables are not eliminated if it produces a resolvent with a length above this limit.
					// -1 means no limit.
	int subsumption_lim; // Do not check if subsumption against a clause larger than this. -1 means no limit.
	double simp_garbage_frac; // A different limit for when to issue a GC during simplification (Also see 'garbage_frac').
	bool use_elim;          // Perform variable elimination.

    // Statistics:
    int     merges;
    int     eliminated_vars;

    // Helper structures:
    //
	struct ElimLt {
		const vec<int> &n_occ;
		explicit ElimLt(const vec<int> &no) :
				n_occ(no) {
		}
		uint64_t cost(Var x) const {
			return (uint64_t) n_occ[toInt(mkLit(x))]
					* (uint64_t) n_occ[toInt(~mkLit(x))];
		}
		bool operator()(Var x, Var y) const {
			return cost(x) < cost(y);
		}
	};

	struct ClauseDeleted {
		const ClauseAllocator &ca;
		explicit ClauseDeleted(const ClauseAllocator &_ca) :
				ca(_ca) {
		}
		bool operator()(const CRef &cr) const {
			return ca[cr].mark() == 1;
		}
	};

    // Solver state:
    //
    int                 elimorder;
    bool                use_simplification;
    vec<uint32_t>       elimclauses;
    vec<char>           touched;
    OccLists<Var, vec<CRef>, ClauseDeleted>
                        occurs;
    vec<int>            n_occ;
    Heap<ElimLt>        elim_heap;
    Queue<CRef>         subsumption_queue;
    vec<char>           frozen;
    vec<char>           eliminated;
    int                 bwdsub_assigns;
    int                 n_touched;

    // Temporaries:
    //
    CRef                bwdsub_tmpunit;

    // Main internal methods:
    //
    void          updateElimHeap           (Var v);
    void          gatherTouchedClauses     ();
    bool          compute_resolvent                    (const Clause& _ps, const Clause& _qs, Var v, vec<Lit>& out_clause);
    bool          count_literals_in_resolvent                    (const Clause& _ps, const Clause& _qs, Var v, int& size);
    bool          backwardSubsumptionCheck (bool verbose = false);
    bool		  elimination_is_desired(Var& v, const vec<CRef>& cls, vec<CRef>& pos,vec<CRef>& neg);
    bool          eliminateVar             (Var v);
    void          extendModel              ();

    void          removeClause             (CRef cr);
    bool          strengthenClause         (CRef cr, Lit l);
    void          relocAll                 (ClauseAllocator& to);

    /*
     * I do not really understand why the following is necessary. Essentially,
     * it cleans up as if after eliminate.
     */
    void initialize_if_no_pre();

private:
    /*
     * I do not understand why the contents of the following function are
     * necessary.
     */
	void cleanup_after_eliminate();
	void cleanup_after_eliminate_(int trail_size_last);
	bool can_be_eliminated_(Var elim);
	bool subsumption_check_is_desired();
	bool elimination_is_desired(const vec<CRef> &pos, const vec<CRef> &neg,
			Var v, const vec<CRef> &cls);
	Var delete_and_store_to_be_resolved_clauses(Var v, const vec<CRef> &pos,
			const vec<CRef> &neg);
	void compute_resolvents(const vec<CRef> &pos, const vec<CRef> &neg, Var v);
	void cleanup_after_Variable_elimination(const vec<CRef> &cls, Var v);
	Var var_with_least_occurences(Clause &c);
};


//=================================================================================================
// Implementation of inline methods:


inline bool SimpSolver::isEliminated (Var v) const { return eliminated[v]; }
inline void SimpSolver::updateElimHeap(Var v) {
    assert(use_simplification);
    // if (!frozen[v] && !isEliminated(v) && value(v) == l_Undef)
    if (elim_heap.inHeap(v) || (!frozen[v] && !isEliminated(v) && variable_database.value(v) == l_Undef))
        elim_heap.update(v); }


//=================================================================================================
}

#endif
