/*****************************************************************************************[Main.cc]
Copyright (c) 2003-2006, Niklas Een, Niklas Sorensson
Copyright (c) 2007,      Niklas Sorensson

 Chanseok Oh's MiniSat Patch Series -- Copyright (c) 2015, Chanseok Oh
 
Maple_LCM, Based on MapleCOMSPS_DRUP -- Copyright (c) 2017, Mao Luo, Chu-Min LI, Fan Xiao: implementing a learnt clause minimisation approach
Reference: M. Luo, C.-M. Li, F. Xiao, F. Manya, and Z. L. , “An effective learnt clause minimization approach for cdcl sat solvers,” in IJCAI-2017, 2017, pp. to–appear.
 
Maple_LCM_Dist, Based on Maple_LCM -- Copyright (c) 2017, Fan Xiao, Chu-Min LI, Mao Luo: using a new branching heuristic called Distance at the beginning of search
 
 
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/

#include <zlib.h>
#include <cstdio>
#include <cstdlib>

#include "../CleanSolver/SimpSolver/SimpSolver.h"
#include "../Utilities/Allocators/XAlloc.h"
#include "../Utilities/BooleanTypes/SolverTypes.h"
#include "../Utilities/ParseUtilities/DimacsReader/Dimacs.h"
#include "../Utilities/ParseUtilities/OptionsParser/Options.h"
#include "../Utilities/PrintUtilities/DratPrinter/RefactorHelperDratPrinter.h"
#include "../Utilities/PrintUtilities/LogfilePrinter/LogfilePrinter.h"
#include "../Utilities/PrintUtilities/ResultPrinter/ResultPrinter.h"
#include "OptionsForMain.h"

using namespace Minisat;

//=================================================================================================

// The following is a hack in order to parse the original file again in the end
// in order to check the solution if a model is found.
static char* original_file_name;

/*
 * This function assumes the dimacs file is given by argv[1].
 */
void parse_Dimacs(int &argc, char **argv,
		Minisat::SimpSolver &S) {
    // Currently, It is necessary to set the parsing paramter during parsing
    // as adding the clause would otherwise write the clause into the drat
    // proof, unneccesarrily:
    S.parsing = true;
	if (argc == 1)
		printf("c Reading from standard input... Use '--help' for help.\n");

	gzFile in = (argc == 1) ? gzdopen(0, "rb") : gzopen(argv[1], "rb");
	if (in == NULL)
		printf("c ERROR! Could not open file: %s\n",
				argc == 1 ? "<stdin>" : argv[1]), exit(1);
	else if (argc != 1)
		original_file_name = argv[1];
	parse_DIMACS(in, S);
	gzclose(in);


	print_after_parsing_Dimacs(S);
    S.parsing = false;

}

void initialize_globally() {

	print_initial();
    set_usage_help_message();

	// Disable buffering when run in pipe:
	setbuf(stdout,0);
}


void handle_unsat_after_simplification(Minisat::SimpSolver &S) {
	print_result_if_unsat();
	print_logfile_unsat_after_simplification(S);
	print_drat_proof_if_unsat(S);
    exit(20);
}


void run_the_preprocessor(Minisat::SimpSolver &S){
    S.eliminate(true);
    print_after_preprocessor(S);
    if (!S.ok){
    	handle_unsat_after_simplification(S);
    }

   	handle_special_option_writing_dimacs_after_simplification(S);
}

void check_model(Minisat::SimpSolver& S) {
	gzFile input_stream = gzopen(original_file_name, "rb");
    StreamBuffer in(input_stream);
	check_solution_DIMACS_main(in,S);
}

void handle_if_model_is_found(Minisat::SimpSolver &S) {
	S.extendModel();
	print_logfile_if_model_is_found(S);
	check_model(S);
	print_result_if_model_is_found(S);
	close_drat_proof(S);
}

void handle_if_unsat(Minisat::SimpSolver &S) {
	print_logfile_if_unsat(S);
	print_result_if_unsat();
	print_drat_proof_if_unsat(S);
}


void handle_if_unsolved(const Minisat::SimpSolver &S) {
	print_logfile_if_unsolved(S);
	print_result_if_unsolved();
	close_drat_proof(S);
}

lbool run_the_solver(Minisat::SimpSolver &S) {
	lbool result = S.Solver::solve_();

	if (result == l_True) {
		handle_if_model_is_found(S);
	} else if (result == l_False) {
		handle_if_unsat(S);
	} else {
		handle_if_unsolved(S);
	}

	return result;

}

int fancy_result(lbool result) {
#ifdef NDEBUG
	 	// (faster than "return", which will invoke the destructor for 'Solver')
        exit(result == l_True ? 10 : result == l_False ? 20 : 0);
#else
	return (result == l_True ? 10 : result == l_False ? 20 : 0);
#endif
}

void handle_out_of_memory_exception() {
	printf(
			"c ===============================================================================\n");
	printf("c Out of memory\n");
	printf("s UNKNOWN\n");
    exit(0);
}

//=================================================================================================
// Main:

int main(int argc, char** argv)
{
    try {
    	initialize_globally();

        parseOptions(argc, argv, true);
        set_system_options_from_Main();
        
        SimpSolver  S;
        set_Solver_options_from_Main(S);

        parse_Dimacs(argc,argv,S);

		run_the_preprocessor(S);

		lbool result = run_the_solver(S);

		return fancy_result(result);

        
    } catch (OutOfMemoryException&){
		handle_out_of_memory_exception();
    }
}
