#include "OptionsForMain.h"

#include <signal.h>
#include <fpu_control.h>
#include <sys/resource.h>
#include <unistd.h>
#include <csignal>
#include <cstdint>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#include "../Utilities/PrintUtilities/LogfilePrinter/LogfilePrinter.h"


Minisat::IntOption		verb   ("MAIN", "verb",   "Verbosity level (0=silent, 1=some, 2=more).", 1, Minisat::IntRange(0, 2));
Minisat::BoolOption		pre    ("MAIN", "pre",    "Completely turn on/off any preprocessing.", true);
Minisat::StringOption	dimacs ("MAIN", "dimacs", "If given, stop after preprocessing and write the result to this file.");
Minisat::IntOption		cpu_lim("MAIN", "cpu-lim","Limit on CPU time allowed in seconds.\n", INT32_MAX, Minisat::IntRange(0, INT32_MAX));
Minisat::IntOption		mem_lim("MAIN", "mem-lim","Limit on memory usage in megabytes.\n", INT32_MAX, Minisat::IntRange(0, INT32_MAX));
Minisat::BoolOption		drup   ("MAIN", "drup",   "Generate DRUP UNSAT proof.", false);
Minisat::StringOption	drup_file("MAIN", "drup-file", "DRUP UNSAT proof ouput file.", "");

void set_usage_help_message() {
	Minisat::setUsageHelp(
			"USAGE: %s [options] <input-file> <result-output-file>\n\n  where input may be either in plain or gzipped DIMACS.\n");
}

void set_fpu() {
	fpu_control_t oldcw, newcw;
	_FPU_GETCW(oldcw);
	newcw = (oldcw & ~_FPU_EXTENDED) | _FPU_DOUBLE;
	_FPU_SETCW(newcw);
	printf(
			"c WARNING: for repeatability, setting FPU to use double precision\n");
}

void set_CPU_limit() {
	// Set limit on CPU-time:
	if (cpu_lim != INT32_MAX) {
		rlimit rl;
		getrlimit(RLIMIT_CPU, &rl);
		if (rl.rlim_max == RLIM_INFINITY || (rlim_t) cpu_lim < rl.rlim_max) {
			rl.rlim_cur = cpu_lim;
			if (setrlimit(RLIMIT_CPU, &rl) == -1)
				printf("c WARNING! Could not set resource limit: CPU-time.\n");
		}
	}
}

void set_Memory_limit() {
	// Set limit on virtual memory:
	if (mem_lim != INT32_MAX) {
		rlim_t new_mem_lim = (rlim_t) mem_lim * 1024 * 1024;
		rlimit rl;
		getrlimit(RLIMIT_AS, &rl);
		if (rl.rlim_max == RLIM_INFINITY || new_mem_lim < rl.rlim_max) {
			rl.rlim_cur = new_mem_lim;
			if (setrlimit(RLIMIT_AS, &rl) == -1)
				printf(
						"c WARNING! Could not set resource limit: Virtual memory.\n");
		}
	}
}

void set_system_options_from_Main() {
	set_CPU_limit();
	set_Memory_limit();
}

// Note that '_exit()' rather than 'exit()' has to be used. The reason is that 'exit()' calls
// destructors and may cause deadlocks if a malloc/free function happens to be running (these
// functions are guarded by locks for multithreaded use).
static void SIGINT_exit(int signum) {
	printf("\n");
	printf("c *** INTERRUPTED ***\n");
	if (solver->verbosity > 0) {
		printStats(*solver);
		printf("\n");
		printf("c *** INTERRUPTED ***\n");
	}
	_exit(1);
}

void set_up_signal_handlers(Minisat::SimpSolver &S) {
	// Setting the static solver pointer above in order to be used by the
	// exit function called by the signal handler.
	solver = &S;
	// Use signal handlers that forcibly quit until the solver will be able to respond to
	// interrupts:
	signal(SIGINT, SIGINT_exit);
	signal(SIGXCPU, SIGINT_exit);
	signal(SIGTERM,SIGINT_exit); // For the default send by timeout.
}

void open_file_for_Drat_proof_generation(Minisat::SimpSolver &S) {
	S.drat_printer.drup_file = NULL;
	if (drup || strlen(drup_file)) {
		S.drat_printer.drup_file = strlen(drup_file) ? fopen(drup_file, "wb") : stdout;
		if (S.drat_printer.drup_file == NULL) {
			S.drat_printer.drup_file = stdout;
			printf("c Error opening %s for write.\n", (const char*) drup_file);
		}
		printf("c DRUP proof generation: %s\n",
				S.drat_printer.drup_file == stdout ? "stdout" : drup_file);
	}
}

void set_Solver_options_from_Main(Minisat::SimpSolver &S) {
	S.verbosity = verb;
	set_up_signal_handlers(S);
	open_file_for_Drat_proof_generation(S);
	// The following is supposed to be replaced. FIXME
    if (!pre) S.initialize_if_no_pre();
}

void handle_special_option_writing_dimacs_after_simplification(
		Minisat::SimpSolver &S) {
	if (dimacs) {
		if (S.verbosity > 0)
			printf(
					"c ==============================[ Writing DIMACS ]===============================\n");

		S.toDimacs((const char*) (dimacs));
		if (S.verbosity > 0)
			printStats(S);

		exit(0);
	}
}
