/*****************************************************************************************[Main.cc]

**************************************************************************************************/

#include <errno.h>

#include <signal.h>
#include <zlib.h>
#include <sys/resource.h>

#include <sys/stat.h>

#include "utils/System.h"
#include "utils/ParseUtils.h"
#include "utils/Options.h"
#include "core/Dimacs.h"
#include "simp/SimpSolver.h"

char* input_file = NULL;

using namespace optsat;

FILE*  certifiedOutput;
bool   certifiedUNSAT;   // tarce back UNSAT proof
static const char* _certified = "CORE -- CERTIFIED UNSAT";

//=================================================================================================
#ifndef NUNLOCKED
#define A_putc_unlocked putc_unlocked
#define A_getc_unlocked getc_unlocked
#else
#define A_putc_unlocked putc
#define A_getc_unlocked getc
#endif

// These are signatures for supported compressed file types.  In 2018 the
// SAT Competition was running on StarExec and used internally 'bzip2'
// compressed files, but gave them uncompressed to the solver using exactly
// the same path (with '.bz2' suffix). 

static int xzsig[] = { 0xFD, 0x37, 0x7A, 0x58, 0x5A, 0x00, 0x00, EOF };
static int bz2sig[] = { 0x42, 0x5A, 0x68, EOF };
static int gzsig[] = { 0x1F, 0x8B, EOF };
static int zipsig[] = {0x50, 0x4B, 0x03, 0x04, EOF };
static int sig7z[] = { 0x37, 0x7A, 0xBC, 0xAF, 0x27, 0x1C, EOF };
static int lzmasig[] = { 0x5D, 0x00, 0x00, 0x80, 0x00, EOF };
static int pipefile=true;

bool match (const char * path, const int * sig) 
{
  FILE * tmp = fopen (path, "r");
  if (!tmp) {
    printf("c WARNING: failed to open '%s' to check signature \n", path);
    return false;
  }
  bool res = true;
  for (const int *p = sig; res && (*p != EOF); p++) res = (A_getc_unlocked (tmp) == *p);
  fclose (tmp);
  if (!res) printf("c file type signature check for '%s' failed\n", path);
  else printf("c signature check for '%s' success\n", path);
  return res;
}

bool exists (const char * path) 
{
  struct stat buf;
  if (stat (path, &buf)) return false;
  if (access (path, R_OK)) return false;
  return true;
}

char * find (const char * prg) 
{
  size_t prglen = strlen (prg);
  const char * c = getenv ("PATH");
  if (!c) return 0;;
  size_t len = strlen (c);
  char * e = new char[len + 1];
  strcpy (e, c);
  char * res = 0;
  for (char * p = e, * q; !res && p < e + len; p = q) {
    for (q = p; *q && *q != ':'; q++)
      ;
    *q++ = 0;
    size_t pathlen = (q - p) + prglen;
    char * path = new char [pathlen + 1];
    sprintf (path, "%s/%s", p, prg);
    assert (strlen (path) == pathlen);
    if (exists (path)) res = path;
    else delete [] path;
  }
  delete [] e;
  return res;
}

FILE * open_pipe (const char * fmt, const char * path) 
{
  size_t prglen = 0;
  while (fmt[prglen] && fmt[prglen] != ' ') prglen++;
  char * prg = new char [prglen + 1];
  strncpy (prg, fmt, prglen);
  prg[prglen] = 0;
  char * found = find (prg);
  if (found) printf("c found '%s' in path for '%s'\n", found, prg);
  if (!found) printf ("did not find '%s' in path\n", prg);
  delete [] prg;
  if (!found) return 0;
  delete [] found;
  char * cmd = new char [strlen (fmt) + strlen (path)];
  sprintf (cmd, fmt, path);

  printf("c %s \n",cmd);

  FILE * res = popen (cmd, "r");
  delete [] cmd;
  return res;
}

bool has_suffix (const char * str, const char * suffix) {
  size_t k = strlen (str), l = strlen (suffix);
  return k > l && !strcmp (str + k - l, suffix);
}

FILE * check_popen (const char * path) {
  FILE * file=0;
  if (has_suffix (path, ".xz")) {
  //  file = read_pipe ("xz -c -d %s", xzsig, path);
       if(match(path, xzsig)) file=open_pipe ("xz -c -d %s", path); 
  } else if (has_suffix (path, ".lzma")) {
    //file = read_pipe ("lzma -c -d %s", lzmasig, path);
       if(match(path, lzmasig)) file=open_pipe ("lzma -c -d %s", path); 
  } else if (has_suffix (path, ".bz2")) {
    //file = read_pipe ("bzip2 -c -d %s", bz2sig, path);
     if(match(path, bz2sig)) file=open_pipe ("bzip2 -c -d %s", path); 
 } else if (has_suffix (path, ".gz")) {
    //file = read_pipe ("gzip -c -d %s", gzsig, path);
     if(match(path, gzsig)) file=open_pipe ("gzip -c -d %s", path); 
 } else if (has_suffix (path, ".zip")) {
     if(match(path, zipsig)) file=open_pipe ("unzip -p %s", path); 
 } else if (has_suffix (path, ".7z")) {
   // file = read_pipe ("7z x -so %s 2>/dev/null", sig7z, path);
     if(match(path, sig7z)) file=open_pipe ("7z x -so %s 2>/dev/null", path); 
  }
  pipefile=true; 
  if(file==0) {
       file = fopen(path,"r");
       pipefile=false;
  }
  return file;
}

void printStats(Solver& solver)
{
    double cpu_time = cpuTime();
    double mem_used = memUsedPeak();
    printf("c restarts              : %" PRIu64"\n", solver.starts);
    printf("c conflicts             : %-12" PRIu64"   (%.0f /sec)\n", solver.conflicts   , solver.conflicts   /cpu_time);
    printf("c decisions             : %-12" PRIu64"   (%4.2f %% random) (%.0f /sec)\n", solver.decisions, (float)solver.rnd_decisions*100 / (float)solver.decisions, solver.decisions   /cpu_time);
    printf("c propagations          : %-12" PRIu64"   (%.0f /sec)\n", solver.propagations, solver.propagations/cpu_time);
    printf("c conflict literals     : %-12" PRIu64"   (%4.2f %% deleted)\n", solver.tot_literals, (solver.max_literals - solver.tot_literals)*100 / (double)solver.max_literals);
    printf("c backtracks            : %-12" PRIu64"   (NCB %0.f%% , CB %0.f%%)\n", solver.non_chrono_backtrack + solver.chrono_backtrack, (solver.non_chrono_backtrack * 100) / (double)(solver.non_chrono_backtrack + solver.chrono_backtrack), (solver.chrono_backtrack * 100) / (double)(solver.non_chrono_backtrack + solver.chrono_backtrack));
    if (mem_used != 0) printf("c Memory used           : %.2f MB\n", mem_used);
    printf("c CPU time              : %g s\n", cpu_time);
}


static Solver* solver;

void verifyModel(char* filename, vec<bool> & true_var)
{
    SimpSolver  S;
   // gzFile in = gzopen(filename, "rb");
    FILE *in = check_popen(filename);
    if (in == NULL) printf("c ERROR! Could not open file: %s\n", filename), exit(1);
           
// Parse CNF:
    printf("c final verify filename=%s \n",filename);
  
    parse_DIMACS(in, S);
    //gzclose(in);
    if(pipefile) pclose(in);
    else fclose(in);
  
// Check satisfaction:
   vec<CRef> & cs = S.clauses;
  
   for (int i = 0; i < cs.size(); i++){
         Clause& c = S.ca[cs[i]];
        for (int j = 0; j < c.size(); j++){
              if (sign(c[j])==1 && true_var[var(c[j])]==false ) goto Satisfied;
              if (sign(c[j])==0 && true_var[var(c[j])]==true ) goto Satisfied;
        }
//        printf("s INDETERMINATE\n");
//        printf("c FALSE MODEL!!!\n");
        printf("c {");
        for (int j = 0; j < c.size(); j++){
               if(sign(c[j])) printf("-");
               printf("%d:%d ", var(c[j])+1, true_var[var(c[j])]);
        }
        printf(" }\n");
        exit(0);

      Satisfied:;
    }
}

//static void SIGINT_interrupt(int signum) { solver->interrupt(); }

static void SIGINT_exit(int signum) {
    printf("\n"); printf("c *** INTERRUPTED ***\n");
    if (solver->verbosity > 0){
        printStats(*solver);
        printf("\n"); printf("c *** INTERRUPTED ***\n"); }
    _exit(1); }


//=================================================================================================
// Main:
double initial_time;

int main(int argc, char** argv)
{
    try {
        setUsageHelp("USAGE: %s [options] <input-file> <result-output-file>\n\n  where input may be either in plain or gzipped DIMACS.\n");
        printf("c This is optsat Recursion, March 15,2021 by Jingchao Chen\n");
        
#if defined(__linux__)
        fpu_control_t oldcw, newcw;
        _FPU_GETCW(oldcw); newcw = (oldcw & ~_FPU_EXTENDED) | _FPU_DOUBLE; _FPU_SETCW(newcw);
        printf("c WARNING: for repeatability, setting FPU to use double precision\n");
#endif
        // Extra options:
        //
        IntOption    verb   ("MAIN", "verb",   "Verbosity level (0=silent, 1=some, 2=more).", 0, IntRange(0, 2));
        BoolOption   pre    ("MAIN", "pre",    "Completely turn on/off any preprocessing.", true);
        StringOption dimacs ("MAIN", "dimacs", "If given, stop after preprocessing and write the result to this file.");
        IntOption    cpu_lim("MAIN", "cpu-lim","Limit on CPU time allowed in seconds.\n", INT32_MAX, IntRange(0, INT32_MAX));
        IntOption    mem_lim("MAIN", "mem-lim","Limit on memory usage in megabytes.\n", INT32_MAX, IntRange(0, INT32_MAX));
        BoolOption   opt_certified      (_certified, "certified",    "Certified UNSAT using DRUP format", false);
        StringOption opt_certified_file      (_certified, "certified-output",    "Certified UNSAT output file", "NULL");

        parseOptions(argc, argv, true);
        
        SimpSolver  S;
        double      initial_time = cpuTime();

        if (!pre) S.eliminate(true);

        S.parsing = true;
        S.verbosity = verb;
        certifiedUNSAT = opt_certified;
        if(certifiedUNSAT) {
            if(!strcmp(opt_certified_file,"NULL")) certifiedOutput = stdout;
            else certifiedOutput =  fopen64(opt_certified_file, "wb");	    
        }

        solver = &S;
        // Use signal handlers that forcibly quit until the solver will be able to respond to
        // interrupts:
        signal(SIGINT, SIGINT_exit);
        signal(SIGXCPU,SIGINT_exit);

        // Set limit on CPU-time:
        if (cpu_lim != INT32_MAX){
            rlimit rl;
            getrlimit(RLIMIT_CPU, &rl);
            if (rl.rlim_max == RLIM_INFINITY || (rlim_t)cpu_lim < rl.rlim_max){
                rl.rlim_cur = cpu_lim;
                if (setrlimit(RLIMIT_CPU, &rl) == -1)
                    printf("c WARNING! Could not set resource limit: CPU-time.\n");
            } }

        // Set limit on virtual memory:
        if (mem_lim != INT32_MAX){
            rlim_t new_mem_lim = (rlim_t)mem_lim * 1024*1024;
            rlimit rl;
            getrlimit(RLIMIT_AS, &rl);
            if (rl.rlim_max == RLIM_INFINITY || new_mem_lim < rl.rlim_max){
                rl.rlim_cur = new_mem_lim;
                if (setrlimit(RLIMIT_AS, &rl) == -1)
                    printf("c WARNING! Could not set resource limit: Virtual memory.\n");
            } }
        
        if (argc == 1)
            printf("c Reading from standard input... Use '--help' for help.\n");
        else input_file=argv[1];
 
        //gzFile in = (argc == 1) ? gzdopen(0, "rb") : gzopen(argv[1], "rb");
        FILE *in = (argc == 1) ? stdin : check_popen(argv[1]);
        if (in == NULL)
            printf("c ERROR! Could not open file: %s\n", argc == 1 ? "<stdin>" : argv[1]), exit(1);
        
        if (S.verbosity > 0){
            printf("c ============================[ Problem Statistics ]=============================\n");
            printf("c |                                                                             |\n"); }
        
        parse_DIMACS(in, S);
//        gzclose(in);
        if(pipefile) pclose(in);
        else fclose(in);

        FILE* res = (argc >= 3) ? fopen(argv[2], "wb") : NULL;

        if (S.verbosity > 0){
            printf("c |  Number of variables:  %12d                                         |\n", S.nVars());
            printf("c |  Number of clauses:    %12d                                         |\n", S.nClauses()); }
        
        double parsed_time = initial_time=cpuTime();
        if (S.verbosity > 0)
            printf("c |  Parse time:           %12.2f s                                       |\n", parsed_time - initial_time);

        S.parsing = false;
        S.eliminate(true);
        double simplified_time = cpuTime();
        if (S.verbosity > 0){
            printf("c |  Simplification time:  %12.2f s                                       |\n", simplified_time - parsed_time);
            printf("c |                                                                             |\n"); }

        if (!S.okay()){
            if (res != NULL) fprintf(res, "UNSAT\n"), fclose(res);
            if (S.verbosity > 0){
                printf("c ===============================================================================\n");
                printf("c Solved by simplification\n");
                printStats(S);
                printf("\n"); }
                printf("s UNSATISFIABLE\n");
    
                if (certifiedUNSAT) {
                    fprintf(certifiedOutput, "0\n");
                    if(certifiedOutput != stdout) fclose(certifiedOutput);
                }
                exit(20);
        }

        if (dimacs){
            if (S.verbosity > 0)
                printf("c ==============================[ Writing DIMACS ]===============================\n");
            S.toDimacs((const char*)dimacs);
            if (S.verbosity > 0)
                printStats(S);
            exit(0);
        }

        vec<Lit> dummy;
        lbool ret = S.solveLimited(dummy);
        
        if (S.verbosity > 0){
            printStats(S);
            printf("\n");
        }
         if (ret==l_True && input_file != NULL) {
           vec<bool>   VarValue(S.nVars(), false);
           for (int i = 0; i < S.nVars(); i++){
	          if (S.model[i] == l_True) VarValue[i]=true;
              else VarValue[i]=false;
           }
           if(S.nVars()<5000000){
               verifyModel(input_file, VarValue);
               printf("c OK! verified \n");
           }
        }
 
        printf(ret == l_True ? "s SATISFIABLE\n" : ret == l_False ? "s UNSATISFIABLE\n" : "s UNKNOWN\n");
        if (ret == l_True){
            printf("v ");
            for (int i = 0; i < S.nVars(); i++)
                if (S.model[i] != l_Undef)
                    printf("%s%s%d", (i==0)?"":" ", (S.model[i]==l_True)?"":"-", i+1);
            printf(" 0\n");
        }

        if (certifiedUNSAT && ret == l_False) {
              fprintf(certifiedOutput, "0\n");
              if(certifiedOutput != stdout) fclose(certifiedOutput);
        }
 
        if (res != NULL){
            if (ret == l_True){
                fprintf(res, "SAT\n");
                for (int i = 0; i < S.nVars(); i++)
                    if (S.model[i] != l_Undef)
                        fprintf(res, "%s%s%d", (i==0)?"":" ", (S.model[i]==l_True)?"":"-", i+1);
                fprintf(res, " 0\n");
            }else if (ret == l_False)
                fprintf(res, "UNSAT\n");
            else
                fprintf(res, "INDET\n");
            fclose(res);
        }

#ifdef NDEBUG
        exit(ret == l_True ? 10 : ret == l_False ? 20 : 0);     // (faster than "return", which will invoke the destructor for 'Solver')
#else
        return (ret == l_True ? 10 : ret == l_False ? 20 : 0);
#endif
    } catch (OutOfMemoryException&){
        printf("c ===============================================================================\n");
        printf("c Out of memory\n");
        printf("s UNKNOWN\n");
        exit(0);
    }
}
